from typing import Type

from pydantic import BaseModel


class BaseTool(BaseModel):
    name: str
    arguments: Type[BaseModel]
    description: str = ""
    strict: bool = True

    def to_openai_tool(self):
        schema = self.arguments.model_json_schema()
        schema["additionalProperties"] = False

        if "properties" in schema:
            for prop_name, prop_schema in schema["properties"].items():
                if prop_schema.get("type") == "array":
                    items_schema = prop_schema.get("items", {})
                    if not isinstance(items_schema, dict) or not items_schema.get("type"):
                        prop_schema["items"] = {"type": "string"}

                elif "anyOf" in prop_schema:
                    for sub_schema in prop_schema["anyOf"]:
                        if isinstance(sub_schema, dict) and sub_schema.get("type") == "array":
                            items_schema = sub_schema.get("items", {})
                            if not isinstance(items_schema, dict) or not items_schema.get("type"):
                                sub_schema["items"] = {"type": "string"}

        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": schema,
                "strict": self.strict,
            },
        }

    def to_anthropic_tool(self):
        schema = self.arguments.model_json_schema()
        schema["additionalProperties"] = False

        return {
            "name": self.name,
            "description": self.description,
            "input_schema": {
                "type": "object",
                "properties": schema["properties"],
                "required": schema.get("required", []),
            },
        }

    def to_mistral_tool(self):
        schema = self.arguments.model_json_schema()
        properties = {}
        for prop_name, prop in schema.get("properties", {}).items():
            prop_type = prop["type"]
            if prop_type == "array" and "items" in prop:
                properties[prop_name] = {
                    "type": "array",
                    "items": prop["items"],
                    "description": prop.get("description", ""),
                }
                continue

            properties[prop_name] = {
                "type": prop_type,
                "description": prop.get("description", ""),
            }
            if "enum" in prop:
                properties[prop_name]["enum"] = prop["enum"]

        parameters = {
            "type": "object",
            "properties": properties,
            "required": schema.get("required", []),
            "additionalProperties": False,
        }
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": parameters,
            },
        }

    def to_gemini_tool(self):
        schema = self.arguments.model_json_schema()
        # Convert Pydantic schema types to Gemini schema types
        properties = {}
        for name, prop in schema["properties"].items():
            prop_type = prop.get("type", "string")
            if prop_type == "array" and "items" in prop:
                properties[name] = {
                    "type": "array",
                    "items": prop["items"],
                    "description": prop.get("description", ""),
                }
                continue

            properties[name] = {
                "type": prop_type,
                "description": prop.get("description", ""),
            }
            if "enum" in prop:
                properties[name]["enum"] = prop["enum"]

        return {
            "name": self.name,
            "description": self.description,
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": schema.get("required", []),
            },
        }



import os

from dotenv import load_dotenv
from groq import AsyncGroq, Groq

from synth_ai.zyk.lms.vendors.openai_standard import OpenAIStandard

load_dotenv()


class GroqAPI(OpenAIStandard):
    def __init__(self):
        super().__init__(
            sync_client=Groq(api_key=os.getenv("GROQ_API_KEY")),
            async_client=AsyncGroq(api_key=os.getenv("GROQ_API_KEY")),
        )



import json
import os
import sqlite3
from dataclasses import dataclass
from typing import Optional, Type, Union

from pydantic import BaseModel

from synth_ai.zyk.lms.vendors.base import BaseLMResponse


@dataclass
class PersistentCache:
    def __init__(self, db_path: str = ".cache/persistent_cache.db"):
        os.makedirs(os.path.dirname(db_path), exist_ok=True)
        self.conn = sqlite3.connect(db_path)
        self.cursor = self.conn.cursor()
        self.cursor.execute("""CREATE TABLE IF NOT EXISTS cache
                              (key TEXT PRIMARY KEY, response TEXT)""")
        self.conn.commit()

    def hit_cache(
        self, key: str, response_model: Optional[Type[BaseModel]] = None
    ) -> Optional[BaseLMResponse]:
        self.cursor.execute("SELECT response FROM cache WHERE key = ?", (key,))
        result = self.cursor.fetchone()
        if not result:
            return None

        try:
            cache_data = json.loads(result[0])
        except json.JSONDecodeError:
            # Handle legacy string responses
            return BaseLMResponse(
                raw_response=result[0], structured_output=None, tool_calls=None
            )

        if not isinstance(cache_data, dict):
            return BaseLMResponse(
                raw_response=cache_data, structured_output=None, tool_calls=None
            )

        raw_response = cache_data.get("raw_response")
        tool_calls = cache_data.get("tool_calls")
        structured_output = cache_data.get("structured_output")

        if response_model and structured_output:
            structured_output = response_model(**structured_output)

        return BaseLMResponse(
            raw_response=raw_response,
            structured_output=structured_output,
            tool_calls=tool_calls,
        )

    def add_to_cache(self, key: str, response: Union[BaseLMResponse, str]) -> None:
        if isinstance(response, str):
            cache_data = response
        elif isinstance(response, BaseLMResponse):
            cache_data = {
                "raw_response": response.raw_response
                if response.raw_response is not None
                else None,
                "tool_calls": response.tool_calls
                if response.tool_calls is not None
                else None,
                "structured_output": (
                    response.structured_output.model_dump()
                    if response.structured_output is not None
                    else None
                ),
            }
        else:
            raise ValueError(f"Invalid response type: {type(response)}")

        self.cursor.execute(
            "INSERT OR REPLACE INTO cache (key, response) VALUES (?, ?)",
            (key, json.dumps(cache_data)),
        )
        self.conn.commit()

    def close(self) -> None:
        self.conn.close()



from synth_ai.zyk.lms.caching.handler import CacheHandler

cache_handler = CacheHandler(use_ephemeral_store=True, use_persistent_store=True)
ephemeral_cache_handler = CacheHandler(
    use_ephemeral_store=True, use_persistent_store=False
)


def get_cache_handler(use_ephemeral_cache_only: bool = False):
    if use_ephemeral_cache_only:
        return ephemeral_cache_handler
    else:
        return cache_handler



import os

from together import AsyncTogether, Together

from synth_ai.zyk.lms.vendors.openai_standard import OpenAIStandard


class TogetherAPI(OpenAIStandard):
    def __init__(self):
        self.sync_client = Together(api_key=os.getenv("TOGETHER_API_KEY"))
        self.async_client = AsyncTogether(api_key=os.getenv("TOGETHER_API_KEY"))



import os
from dotenv import load_dotenv


def should_use_cache() -> bool:
    load_dotenv()
    cache_env = os.getenv("USE_ZYK_CACHE", "true").lower()
    return cache_env not in ("false", "0", "no")

reasoning_models = ["o1","o3-mini", "o3", "o4-mini", "claude-3-7-sonnet-latest"]


import re
import os
import json
import asyncio
import time
from typing import Any, Dict, List, Optional, Tuple, Type
import requests
import httpx
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import random
from urllib.parse import urlparse

from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.caching.initialize import get_cache_handler

# Exception types for retry
CUSTOM_ENDPOINT_EXCEPTIONS_TO_RETRY: Tuple[Type[Exception], ...] = (
    requests.RequestException, 
    requests.Timeout,
    httpx.RequestError,
    httpx.TimeoutException
)

class CustomEndpointAPI(VendorBase):
    """Generic vendor client for custom OpenAI-compatible endpoints."""
    
    used_for_structured_outputs: bool = False
    exceptions_to_retry: List = list(CUSTOM_ENDPOINT_EXCEPTIONS_TO_RETRY)
    
    def __init__(self, endpoint_url: str):
        # Validate and sanitize URL
        self._validate_endpoint_url(endpoint_url)
        self.endpoint_url = endpoint_url
        
        # Construct full chat completions URL
        if endpoint_url.endswith('/'):
            endpoint_url = endpoint_url[:-1]
        self.chat_completions_url = f"https://{endpoint_url}/chat/completions"
        self.health_url = f"https://{endpoint_url}/health"
        
        # Setup session with connection pooling and retries
        self.session = self._create_session()
        self.async_client = None  # Lazy init
        
        # Get auth token from environment (generic support for any auth)
        self.auth_token = os.environ.get("CUSTOM_ENDPOINT_API_TOKEN")
        
    def _validate_endpoint_url(self, url: str) -> None:
        """Validate endpoint URL format and prevent SSRF."""
        # Block dangerous URL patterns
        dangerous_patterns = [
            "file://", "ftp://", "gopher://",
            "localhost", "127.", "0.0.0.0",
            "10.", "192.168.", "172.16.", "172.17.", "172.18.", "172.19.",
            "172.20.", "172.21.", "172.22.", "172.23.", "172.24.", "172.25.",
            "172.26.", "172.27.", "172.28.", "172.29.", "172.30.", "172.31.",
            "169.254.",  # link-local
            "::1", "fc00:", "fd00:", "fe80:",  # IPv6 private
        ]
        
        for pattern in dangerous_patterns:
            if pattern in url.lower():
                raise ValueError(f"Blocked URL pattern for security: {pattern}")
                
        # Limit URL length
        if len(url) > 256:
            raise ValueError(f"Endpoint URL too long (max 256 chars)")
            
        # Basic URL format check
        if not re.match(r'^[a-zA-Z0-9\-._~:/?#\[\]@!$&\'()*+,;=]+$', url):
            raise ValueError(f"Invalid URL format: {url}")
    
    def _create_session(self) -> requests.Session:
        """Create session with retry strategy and connection pooling."""
        session = requests.Session()
        
        # Exponential backoff with jitter
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["POST", "GET"]
        )
        
        adapter = HTTPAdapter(
            max_retries=retry_strategy,
            pool_connections=10,
            pool_maxsize=20
        )
        
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        
        return session
    
    async def _get_async_client(self) -> httpx.AsyncClient:
        """Lazy init async client with shared retry logic."""
        if self.async_client is None:
            self.async_client = httpx.AsyncClient(
                timeout=httpx.Timeout(30.0),
                limits=httpx.Limits(max_connections=10, max_keepalive_connections=5)
            )
        return self.async_client
    
    def _get_timeout(self, lm_config: Dict[str, Any]) -> float:
        """Get timeout with per-call override support."""
        return lm_config.get("timeout", 
                           float(os.environ.get("CUSTOM_ENDPOINT_REQUEST_TIMEOUT", "30")))
    
    def _get_temperature_override(self) -> Optional[float]:
        """Get temperature override from environment for this specific endpoint."""
        # Create a safe env var key from the endpoint URL
        # e.g., "example.com/api" -> "CUSTOM_ENDPOINT_TEMP_EXAMPLE_COM_API"
        safe_key = re.sub(r'[^A-Za-z0-9]', '_', self.endpoint_url).upper()
        safe_key = safe_key[:64]  # Limit length
        
        env_key = f"CUSTOM_ENDPOINT_TEMP_{safe_key}"
        temp_str = os.environ.get(env_key)
        return float(temp_str) if temp_str else None
    
    def _compress_tool_schema(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        """Compress JSON schema to reduce token usage."""
        if isinstance(schema, dict):
            # Remove verbose keys
            compressed = {
                k: self._compress_tool_schema(v) 
                for k, v in schema.items() 
                if k not in ["title", "$ref", "$schema"]
            }
            # Shorten descriptions
            if "description" in compressed and len(compressed["description"]) > 50:
                compressed["description"] = compressed["description"][:47] + "..."
            return compressed
        elif isinstance(schema, list):
            return [self._compress_tool_schema(item) for item in schema]
        return schema
    
    def _inject_tools_into_prompt(self, system_message: str, tools: List[BaseTool]) -> str:
        """Inject tool definitions with compressed schemas and clear output format."""
        if not tools:
            return system_message
            
        tool_descriptions = []
        for tool in tools:
            schema = tool.arguments.model_json_schema()
            compressed_schema = self._compress_tool_schema(schema)
            
            tool_desc = f"Tool: {tool.name}\nDesc: {tool.description}\nParams: {json.dumps(compressed_schema, separators=(',', ':'))}"
            tool_descriptions.append(tool_desc)
        
        tools_text = "\n".join(tool_descriptions)
        
        return f"""{system_message}

Available tools:
{tools_text}

IMPORTANT: To use a tool, respond with JSON wrapped in ```json fences:
```json
{{"tool_call": {{"name": "tool_name", "arguments": {{...}}}}}}
```

For regular responses, just respond normally without JSON fences."""

    def _extract_tool_calls(self, content: str, tools: List[BaseTool]) -> tuple[Optional[List], str]:
        """Extract and validate tool calls from response."""
        # Look for JSON fenced blocks
        json_pattern = r'```json\s*(\{.*?\})\s*```'
        matches = re.findall(json_pattern, content, re.DOTALL)
        
        if not matches:
            return None, content
            
        tool_calls = []
        cleaned_content = content
        
        for match in matches:
            try:
                tool_data = json.loads(match)
                if "tool_call" in tool_data:
                    call_data = tool_data["tool_call"]
                    tool_name = call_data.get("name")
                    
                    # Validate against available tools
                    matching_tool = next((t for t in tools if t.name == tool_name), None)
                    if matching_tool:
                        # Validate arguments with pydantic
                        validated_args = matching_tool.arguments(**call_data.get("arguments", {}))
                        tool_calls.append({
                            "name": tool_name,
                            "arguments": validated_args.model_dump()
                        })
                        
                        # Remove tool call from content
                        cleaned_content = cleaned_content.replace(f"```json\n{match}\n```", "").strip()
                        
            except (json.JSONDecodeError, Exception):
                # Fall back to treating as normal text if validation fails
                continue
                
        return tool_calls if tool_calls else None, cleaned_content
    
    def _exponential_backoff_with_jitter(self, attempt: int) -> float:
        """Calculate backoff time with jitter to prevent thundering herd."""
        base_delay = min(2 ** attempt, 32)  # Cap at 32 seconds
        jitter = random.uniform(0, 1)
        return base_delay + jitter
    
    def _handle_rate_limit(self, response: requests.Response) -> None:
        """Extract and propagate rate limit information."""
        if response.status_code == 429:
            retry_after = response.headers.get("Retry-After")
            if retry_after:
                # Bubble up to synth-ai scheduler
                raise requests.exceptions.RetryError(f"Rate limited. Retry after {retry_after}s")
    
    async def _hit_api_async(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "low",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        """Async API call with comprehensive error handling and streaming support."""
        
        # Cache integration - check first
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result
        
        # Apply tool injection
        if tools and messages:
            messages = messages.copy()
            if messages and messages[0].get("role") == "system":
                messages[0]["content"] = self._inject_tools_into_prompt(
                    messages[0]["content"], tools
                )
        
        # Prepare request
        headers = {"Content-Type": "application/json"}
        if self.auth_token:
            headers["Authorization"] = f"Bearer {self.auth_token}"
        
        # Apply temperature override
        temp_override = self._get_temperature_override()
        request_temp = temp_override if temp_override else lm_config.get("temperature", 0.7)
        
        payload = {
            "model": model,  # Pass through the model name
            "messages": messages,
            "temperature": request_temp,
            "stream": lm_config.get("stream", False)
        }
        
        timeout = self._get_timeout(lm_config)
        client = await self._get_async_client()
        
        # Make request with retry logic
        for attempt in range(3):
            try:
                response = await client.post(
                    self.chat_completions_url,
                    json=payload,
                    headers=headers,
                    timeout=timeout
                )
                
                if response.status_code == 429:
                    self._handle_rate_limit(response)
                    
                response.raise_for_status()
                
                response_data = response.json()
                content = response_data["choices"][0]["message"]["content"]
                
                # Extract tool calls
                tool_calls, clean_content = self._extract_tool_calls(content, tools or [])
                
                lm_response = BaseLMResponse(
                    raw_response=clean_content,
                    structured_output=None,
                    tool_calls=tool_calls
                )
                
                # Add to cache
                used_cache_handler.add_to_managed_cache(
                    model, messages, lm_config=lm_config, output=lm_response, tools=tools
                )
                
                return lm_response
                    
            except (httpx.RequestError, httpx.TimeoutException) as e:
                if attempt == 2:  # Last attempt
                    raise
                await asyncio.sleep(self._exponential_backoff_with_jitter(attempt))
                
    def _hit_api_sync(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "low",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        """Sync version with same logic as async."""
        
        # Cache integration - check first
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result
        
        # Apply tool injection
        if tools and messages:
            messages = messages.copy()
            if messages and messages[0].get("role") == "system":
                messages[0]["content"] = self._inject_tools_into_prompt(
                    messages[0]["content"], tools
                )
        
        # Prepare request
        headers = {"Content-Type": "application/json"}
        if self.auth_token:
            headers["Authorization"] = f"Bearer {self.auth_token}"
        
        # Apply temperature override
        temp_override = self._get_temperature_override()
        request_temp = temp_override if temp_override else lm_config.get("temperature", 0.7)
        
        payload = {
            "model": model,  # Pass through the model name
            "messages": messages,
            "temperature": request_temp,
            "stream": lm_config.get("stream", False)
        }
        
        timeout = self._get_timeout(lm_config)
        
        # Make request with retry logic
        for attempt in range(3):
            try:
                response = self.session.post(
                    self.chat_completions_url,
                    json=payload,
                    headers=headers,
                    timeout=timeout
                )
                
                if response.status_code == 429:
                    self._handle_rate_limit(response)
                    
                response.raise_for_status()
                
                response_data = response.json()
                content = response_data["choices"][0]["message"]["content"]
                
                # Extract tool calls
                tool_calls, clean_content = self._extract_tool_calls(content, tools or [])
                
                lm_response = BaseLMResponse(
                    raw_response=clean_content,
                    structured_output=None,
                    tool_calls=tool_calls
                )
                
                # Add to cache
                used_cache_handler.add_to_managed_cache(
                    model, messages, lm_config=lm_config, output=lm_response, tools=tools
                )
                
                return lm_response
                    
            except (requests.RequestException, requests.Timeout) as e:
                if attempt == 2:  # Last attempt
                    raise
                time.sleep(self._exponential_backoff_with_jitter(attempt))
    
    def __del__(self):
        """Cleanup resources."""
        if hasattr(self, 'session'):
            self.session.close()
        if hasattr(self, 'async_client') and self.async_client:
            # Schedule cleanup for async client
            pass


from synth_ai.zyk.lms.vendors.core.anthropic_api import AnthropicAPI
from synth_ai.zyk.lms.vendors.core.gemini_api import GeminiAPI
from synth_ai.zyk.lms.vendors.core.openai_api import (
    OpenAIPrivate,
    OpenAIStructuredOutputClient,
)
from synth_ai.zyk.lms.vendors.supported.deepseek import DeepSeekAPI
from synth_ai.zyk.lms.vendors.supported.together import TogetherAPI
from synth_ai.zyk.lms.vendors.supported.groq import GroqAPI
from synth_ai.zyk.lms.vendors.core.mistral_api import MistralAPI
from synth_ai.zyk.lms.vendors.supported.custom_endpoint import CustomEndpointAPI


class OpenAIClient(OpenAIPrivate):
    def __init__(self, synth_logging: bool = True):
        super().__init__(
            synth_logging=synth_logging,
        )


class AnthropicClient(AnthropicAPI):
    def __init__(self):
        super().__init__()


class GeminiClient(GeminiAPI):
    def __init__(self):
        super().__init__()


class DeepSeekClient(DeepSeekAPI):
    def __init__(self):
        super().__init__()


class TogetherClient(TogetherAPI):
    def __init__(self):
        super().__init__()


class GroqClient(GroqAPI):
    def __init__(self):
        super().__init__()


class MistralClient(MistralAPI):
    def __init__(self):
        super().__init__()


class CustomEndpointClient(CustomEndpointAPI):
    def __init__(self, endpoint_url: str):
        super().__init__(endpoint_url=endpoint_url)



import hashlib
from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel

from synth_ai.zyk.lms.caching.ephemeral import EphemeralCache
from synth_ai.zyk.lms.caching.persistent import PersistentCache
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse

persistent_cache = PersistentCache()
ephemeral_cache = EphemeralCache()

import logging
logger = logging.getLogger(__name__)

def map_params_to_key(
    messages: List[Dict],
    model: str,
    temperature: float,
    response_model: Optional[Type[BaseModel]],
    tools: Optional[List[BaseTool]] = None,
    reasoning_effort: str = "low",
) -> str:
    if any(m is None for m in messages):
        raise ValueError("Messages cannot contain None values - messages: ", messages)
    if not all([isinstance(msg["content"], str) for msg in messages]):
        normalized_messages = "".join([str(msg["content"]) for msg in messages])
    else:
        normalized_messages = "".join([msg["content"] for msg in messages])
    normalized_model = model
    normalized_temperature = f"{temperature:.2f}"[:4]
    normalized_response_model = str(response_model.schema()) if response_model else ""
    normalized_reasoning_effort = reasoning_effort if reasoning_effort else ""

    # Normalize tools if present
    normalized_tools = ""
    if tools and all(isinstance(tool, BaseTool) for tool in tools):
        tool_schemas = []
        for tool in sorted(tools, key=lambda x: x.name):  # Sort by name for consistency
            tool_schema = {
                "name": tool.name,
                "description": tool.description,
                "arguments": tool.arguments.schema(),
            }
            tool_schemas.append(str(tool_schema))
        #logger.error(f"Tool schemas: {tool_schemas}")
        normalized_tools = "".join(tool_schemas)
    elif tools:
        #logger.error(f"Tools: {tools}")
        normalized_tools = "".join([str(tool) for tool in tools])

    key_str = ""
    components = [
        normalized_messages,
        normalized_model,
        normalized_temperature,
        normalized_response_model,
        normalized_tools,
        normalized_reasoning_effort
    ]
    for component in components:
        if component:
            key_str += str(component)

    return hashlib.sha256(key_str.encode()).hexdigest() 


class CacheHandler:
    use_persistent_store: bool = False
    use_ephemeral_store: bool = True

    def __init__(
        self, use_persistent_store: bool = False, use_ephemeral_store: bool = True
    ):
        self.use_persistent_store = use_persistent_store
        self.use_ephemeral_store = use_ephemeral_store

    def _validate_messages(self, messages: List[Dict[str, Any]]) -> None:
        """Validate that messages are in the correct format."""
        assert all(
            [type(msg["content"]) == str for msg in messages]
        ), "All message contents must be strings"

    def hit_managed_cache(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        tools: Optional[List[BaseTool]] = None,
    ) -> Optional[BaseLMResponse]:
        """Hit the cache with the given key."""
        self._validate_messages(messages)
        assert type(lm_config) == dict, "lm_config must be a dictionary"
        key = map_params_to_key(
            messages,
            model,
            lm_config.get("temperature", 0.0),
            lm_config.get("response_model", None),
            tools,
            lm_config.get("reasoning_effort", "low"),
        )
        if self.use_persistent_store:
            return persistent_cache.hit_cache(
                key=key, response_model=lm_config.get("response_model", None)
            )
        elif self.use_ephemeral_store:
            return ephemeral_cache.hit_cache(
                key=key, response_model=lm_config.get("response_model", None)
            )
        else:
            return None

    def add_to_managed_cache(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        output: BaseLMResponse,
        tools: Optional[List[BaseTool]] = None,
    ) -> None:
        """Add the given output to the cache."""
        self._validate_messages(messages)
        assert type(output) == BaseLMResponse, "output must be a BaseLMResponse"
        assert type(lm_config) == dict, "lm_config must be a dictionary"
        key = map_params_to_key(
            messages,
            model,
            lm_config.get("temperature", 0.0),
            lm_config.get("response_model", None),
            tools,
            lm_config.get("reasoning_effort", "low"),
        )
        if self.use_persistent_store:
            persistent_cache.add_to_cache(key, output)
        if self.use_ephemeral_store:
            ephemeral_cache.add_to_cache(key, output)



import json
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    get_type_hints,
    get_args,
    get_origin,
    Union,
    Literal,
)
from pydantic import BaseModel
import warnings


def generate_type_map() -> Dict[Any, str]:
    base_types = {
        int: "int",
        float: "float",
        bool: "bool",
        str: "str",
        Any: "Any",
    }

    collection_types = {
        List: "List",
        Dict: "Dict",
        Optional: "Optional",
    }

    type_map = {}
    for base_type, name in base_types.items():
        type_map[base_type] = name
        for collection, collection_name in collection_types.items():
            if collection is Optional:
                type_map[Optional[base_type]] = name
            elif collection is Dict:
                # Handle generic Dict type
                type_map[Dict] = "Dict[Any,Any]"
                # Provide both key and value types for Dict
                type_map[Dict[base_type, base_type]] = (
                    f"{collection_name}[{name},{name}]"
                )
                # Handle Dict[Any, Any] explicitly
                type_map[Dict[Any, Any]] = "Dict[Any,Any]"
            else:
                type_map[collection[base_type]] = f"{collection_name}[{name}]"
    return type_map


def generate_example_dict() -> Dict[str, Any]:
    example_values = {
        "str": "<Your type-str response here>",
        "int": "<Your type-int response here>",
        "float": "<Your type-float response here>",
        "bool": "<Your type-bool response here>",
        "Any": "<Your response here (infer the type from context)>",
    }

    example_dict = {}
    for key, value in example_values.items():
        example_dict[key] = value
        example_dict[f"List[{key}]"] = [value]
        example_dict[f"List[List[{key}]]"] = [[value]]
        if key == "Dict[str,str]":
            example_dict[f"Dict[str,{key.split('[')[1]}"] = {value: value}
        elif key.startswith("Dict"):
            example_dict[key] = {value: value}

    # Add example for Dict[Any,Any]
    example_dict["Dict[Any,Any]"] = {"<key>": "<value>"}

    return example_dict


base_type_examples = {
    int: ("int", 42),
    float: ("float", 3.14),
    bool: ("bool", True),
    str: ("str", "example"),
    Any: ("Any", "<Any value>"),
}


def get_type_string(type_hint):
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    if origin is None:
        if isinstance(type_hint, type) and issubclass(type_hint, BaseModel):
            # For Pydantic models, create a dictionary of field types
            field_types = {}
            for field_name, field_info in type_hint.model_fields.items():
                field_type = get_type_string(field_info.annotation)
                # Check for Literal type by looking at the origin
                if get_origin(field_info.annotation) is Literal:
                    literal_args = get_args(field_info.annotation)
                    field_type = f"Literal[{repr(literal_args[0])}]"
                field_types[field_name] = field_type
            return f"{type_hint.__name__}({', '.join(f'{k}: {v}' for k, v in field_types.items())})"
        else:
            return base_type_examples.get(type_hint, ("Unknown", "unknown"))[0]
    elif origin in (list, List):
        elem_type = get_type_string(args[0])
        return f"List[{elem_type}]"
    elif origin in (dict, Dict):
        key_type = get_type_string(args[0])
        value_type = get_type_string(args[1])
        return f"Dict[{key_type}, {value_type}]"
    elif origin is Union:
        non_none_types = [t for t in args if t is not type(None)]
        if len(non_none_types) == 1:
            return f"Optional[{get_type_string(non_none_types[0])}]"
        else:
            # For unions of Pydantic models (like tool calls), show each variant
            union_types = []
            for t in non_none_types:
                if isinstance(t, type) and issubclass(t, BaseModel):
                    # Include discriminator field if present
                    discriminator = None
                    for field_name, field_info in t.model_fields.items():
                        if get_origin(field_info.annotation) is Literal:
                            literal_args = get_args(field_info.annotation)
                            discriminator = f"{field_name}={repr(literal_args[0])}"
                            break
                    type_str = t.__name__
                    if discriminator:
                        type_str += f"({discriminator})"
                    union_types.append(type_str)
                else:
                    union_types.append(get_type_string(t))
            return f"Union[{', '.join(union_types)}]"
    elif origin is Literal:
        # Handle Literal type directly
        return f"Literal[{repr(args[0])}]"
    else:
        return "Any"


def get_example_value(type_hint):
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    if origin is None:
        if isinstance(type_hint, type) and issubclass(type_hint, BaseModel):
            example = {}
            union_docs = []
            for field_name, field_info in type_hint.model_fields.items():
                field_value, field_docs = get_example_value(field_info.annotation)
                if field_docs:
                    union_docs.extend(field_docs)

                # Handle literal fields by checking origin
                if get_origin(field_info.annotation) is Literal:
                    literal_args = get_args(field_info.annotation)
                    field_value = literal_args[0]

                # Include field description if available
                if field_info.description:
                    example[field_name] = {
                        "value": field_value,
                        "description": field_info.description,
                    }
                else:
                    example[field_name] = field_value
            return example, union_docs
        else:
            return base_type_examples.get(type_hint, ("Unknown", "unknown"))[1], []
    elif origin in (list, List):
        value, docs = get_example_value(args[0])
        return [value], docs
    elif origin in (dict, Dict):
        if not args or len(args) < 2:
            warnings.warn(
                f"Dictionary type hint {type_hint} missing type arguments. "
                "Defaulting to Dict[str, Any].",
                UserWarning,
            )
            return {"example_key": "<Any value>"}, []  # Default for Dict[str, Any]
        key_value, key_docs = get_example_value(args[0])
        value_value, value_docs = get_example_value(args[1])
        return {key_value: value_value}, key_docs + value_docs
    elif origin is Union:
        non_none_types = [t for t in args if t is not type(None)]
        # For unions of tool calls, use the first one but preserve the discriminator
        first_type = non_none_types[0]
        union_docs = []

        if all(
            isinstance(t, type) and issubclass(t, BaseModel) for t in non_none_types
        ):
            # Generate examples for all union variants
            for t in non_none_types:
                example = {}
                for field_name, field_info in t.model_fields.items():
                    field_value, _ = get_example_value(field_info.annotation)
                    if get_origin(field_info.annotation) is Literal:
                        literal_args = get_args(field_info.annotation)
                        field_value = literal_args[0]
                    example[field_name] = field_value
                union_docs.append(f"\nExample {t.__name__}:")
                union_docs.append(json.dumps(example, indent=2))

        # Return first type as main example
        if isinstance(first_type, type) and issubclass(first_type, BaseModel):
            example, _ = get_example_value(first_type)
            # Ensure tool_type or other discriminator is preserved
            for field_name, field_info in first_type.model_fields.items():
                if get_origin(field_info.annotation) is Literal:
                    literal_args = get_args(field_info.annotation)
                    if (
                        isinstance(example[field_name], dict)
                        and "value" in example[field_name]
                    ):
                        example[field_name]["value"] = literal_args[0]
                    else:
                        example[field_name] = literal_args[0]
            return example, union_docs
        main_example, docs = get_example_value(first_type)
        return main_example, docs + union_docs
    elif origin is Literal:
        # Handle Literal type directly
        return args[0], []
    else:
        return "<Unknown>", []


def add_json_instructions_to_messages(
    system_message,
    user_message,
    response_model: Optional[Type[BaseModel]] = None,
    previously_failed_error_messages: List[str] = [],
) -> Tuple[str, str]:
    if response_model:
        type_hints = get_type_hints(response_model)
        # print("Type hints", type_hints)
        stringified_fields = {}
        union_docs = []
        for key, type_hint in type_hints.items():
            # print("Key", key, "Type hint", type_hint)
            example_value, docs = get_example_value(type_hint)
            union_docs.extend(docs)
            field = response_model.model_fields[key]  # Updated for Pydantic v2

            # Adjusted for Pydantic v2
            field_description = (
                field.description if hasattr(field, "description") else None
            )

            if field_description:
                stringified_fields[key] = (example_value, field_description)
            else:
                stringified_fields[key] = example_value
        example_json = json.dumps(
            {
                k: v[0] if isinstance(v, tuple) else v
                for k, v in stringified_fields.items()
            },
            indent=4,
        )
        description_comments = "\n".join(
            f"// {k}: {v[1]}"
            for k, v in stringified_fields.items()
            if isinstance(v, tuple)
        )

        # print("Example JSON", example_json)
        # print("Description comments", description_comments)
        # print("Union documentation", "\n".join(union_docs))
        # raise Exception("Stop here")

        system_message += f"""\n\n
Please deliver your response in the following JSON format:
```json
{example_json}
```
{description_comments}
"""
    if len(union_docs) != 0:
        system_message += f"""\n\n
NOTE - the above example included a union type. Here are some additional examples for that union type:
{chr(10).join(union_docs)}
"""
    if len(previously_failed_error_messages) != 0:
        system_message += f"""\n\n
Please take special care to follow the format exactly.
Keep in mind the following:
- Always use double quotes for strings

Here are some error traces from previous attempts:
{previously_failed_error_messages}
"""
    return system_message, user_message


def inject_structured_output_instructions(
    messages: List[Dict[str, str]],
    response_model: Optional[Type[BaseModel]] = None,
    previously_failed_error_messages: List[str] = [],
) -> List[Dict[str, str]]:
    prev_system_message_content = messages[0]["content"]
    prev_user_message_content = messages[1]["content"]
    system_message, user_message = add_json_instructions_to_messages(
        prev_system_message_content,
        prev_user_message_content,
        response_model,
        previously_failed_error_messages,
    )
    messages[0]["content"] = system_message
    messages[1]["content"] = user_message
    return messages



import asyncio
import unittest
from typing import List

from pydantic import BaseModel, Field

from synth_ai.zyk.lms.core.main import LM


# Define example structured output models
class SimpleResponse(BaseModel):
    message: str
    confidence: float


class ComplexResponse(BaseModel):
    title: str
    tags: List[str]
    content: str


class NestedResponse(BaseModel):
    main_category: str
    subcategories: List[str]
    details: SimpleResponse


# Define nested structured output models
class Address(BaseModel):
    street: str
    city: str
    country: str


class PersonalInfo(BaseModel):
    name: str
    age: int
    address: Address


class WorkInfo(BaseModel):
    company: str
    position: str
    years_experience: int


class NestedPersonResponse(BaseModel):
    personal: PersonalInfo
    work: WorkInfo
    skills: List[str]


class ProjectDetails(BaseModel):
    name: str
    description: str
    technologies: List[str]


class NestedPortfolioResponse(BaseModel):
    developer: PersonalInfo
    projects: List[ProjectDetails]
    total_experience: int


class NestedCompanyResponse(BaseModel):
    name: str
    founded: int
    headquarters: Address
    employees: List[PersonalInfo]
    main_products: List[str]


class TestLMStructuredOutputs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Initialize the LM once for all tests
        cls.lm = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.7,
            max_retries="Few",
            structured_output_mode="forced_json",
        )

    def test_sync_simple_response(self):
        result = self.lm.respond_sync(
            system_message="You are a helpful assistant.",
            user_message="Give me a short greeting and your confidence level.",
            response_model=SimpleResponse,
        )
        self.assertIsInstance(result, SimpleResponse)
        self.assertIsInstance(result.message, str)
        self.assertIsInstance(result.confidence, float)
        self.assertGreaterEqual(result.confidence, 0)
        self.assertLessEqual(result.confidence, 1)

    def test_sync_complex_response(self):
        result = self.lm.respond_sync(
            system_message="You are a content creator.",
            user_message="Create a short blog post about AI.",
            response_model=ComplexResponse,
        )
        self.assertIsInstance(result, ComplexResponse)
        self.assertIsInstance(result.title, str)
        self.assertIsInstance(result.tags, list)
        self.assertIsInstance(result.content, str)

    async def async_nested_response(self):
        result = await self.lm.respond_async(
            system_message="You are a categorization expert.",
            user_message="Categorize 'Python' and provide a brief description.",
            response_model=NestedResponse,
        )
        self.assertIsInstance(result, NestedResponse)
        self.assertIsInstance(result.main_category, str)
        self.assertIsInstance(result.subcategories, list)
        self.assertIsInstance(result.details, SimpleResponse)

    def test_async_nested_response(self):
        asyncio.run(self.async_nested_response())


class TestLMNestedStructuredOutputs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Initialize the LM once for all tests
        cls.lm = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.7,
            max_retries="Few",
            structured_output_mode="forced_json",
        )

    def test_sync_nested_person_response(self):
        result = self.lm.respond_sync(
            system_message="You are an HR assistant.",
            user_message="Provide detailed information about a fictional employee named John Doe.",
            response_model=NestedPersonResponse,
        )
        self.assertIsInstance(result, NestedPersonResponse)
        self.assertIsInstance(result.personal, PersonalInfo)
        self.assertIsInstance(result.personal.address, Address)
        self.assertIsInstance(result.work, WorkInfo)
        self.assertIsInstance(result.skills, list)

    def test_sync_nested_portfolio_response(self):
        result = self.lm.respond_sync(
            system_message="You are a portfolio manager.",
            user_message="Create a portfolio for a fictional software developer with multiple projects.",
            response_model=NestedPortfolioResponse,
        )
        self.assertIsInstance(result, NestedPortfolioResponse)
        self.assertIsInstance(result.developer, PersonalInfo)
        self.assertIsInstance(result.developer.address, Address)
        self.assertIsInstance(result.projects, list)
        for project in result.projects:
            self.assertIsInstance(project, ProjectDetails)
        self.assertIsInstance(result.total_experience, int)

    async def async_nested_company_response(self):
        result = await self.lm.respond_async(
            system_message="You are a company information specialist.",
            user_message="Provide detailed information about a fictional tech company.",
            response_model=NestedCompanyResponse,
        )
        self.assertIsInstance(result, NestedCompanyResponse)
        self.assertIsInstance(result.headquarters, Address)
        self.assertIsInstance(result.employees, list)
        for employee in result.employees:
            self.assertIsInstance(employee, PersonalInfo)
            self.assertIsInstance(employee.address, Address)
        self.assertIsInstance(result.main_products, list)

    def test_async_nested_company_response(self):
        asyncio.run(self.async_nested_company_response())


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class BaseLMResponse(BaseModel):
    raw_response: str
    structured_output: Optional[BaseModel] = None
    tool_calls: Optional[List[Dict]] = None


class VendorBase(ABC):
    used_for_structured_outputs: bool = False
    exceptions_to_retry: List[Exception] = []

    @abstractmethod
    async def _hit_api_async(
        self,
        messages: List[Dict[str, Any]],
        response_model_override: Optional[BaseModel] = None,
    ) -> str:
        pass

    @abstractmethod
    def _hit_api_sync(
        self,
        messages: List[Dict[str, Any]],
        response_model_override: Optional[BaseModel] = None,
    ) -> str:
        pass



import json
from typing import Any, Dict, List, Optional, Tuple, Type

import openai
import pydantic_core

# from openai import AsyncOpenAI, OpenAI
from pydantic import BaseModel

from synth_ai.zyk.lms.caching.initialize import get_cache_handler
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse
from synth_ai.zyk.lms.constants import SPECIAL_BASE_TEMPS, OPENAI_REASONING_MODELS
from synth_ai.zyk.lms.vendors.openai_standard import OpenAIStandard

OPENAI_EXCEPTIONS_TO_RETRY: Tuple[Type[Exception], ...] = (
    pydantic_core._pydantic_core.ValidationError,
    openai.OpenAIError,
    openai.APIConnectionError,
    openai.RateLimitError,
    openai.APIError,
    openai.Timeout,
    openai.InternalServerError,
    openai.APIConnectionError,
)


class OpenAIStructuredOutputClient(OpenAIStandard):
    def __init__(self, synth_logging: bool = True):
        if synth_logging:
            # print("Using synth logging - OpenAIStructuredOutputClient")
            from synth_sdk import AsyncOpenAI, OpenAI
        else:
            # print("Not using synth logging - OpenAIStructuredOutputClient")
            from openai import AsyncOpenAI, OpenAI

        super().__init__(
            used_for_structured_outputs=True,
            exceptions_to_retry=OPENAI_EXCEPTIONS_TO_RETRY,
            sync_client=OpenAI(),
            async_client=AsyncOpenAI(),
        )

    async def _hit_api_async_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        tools: Optional[List[BaseTool]] = None,
        reasoning_effort: str = "high",
    ) -> str:
        if tools:
            raise ValueError("Tools are not supported for async structured output")
        # "Hit client")
        lm_config = {"temperature": temperature, "response_model": response_model, "reasoning_effort": reasoning_effort}
        used_cache_handler = get_cache_handler(
            use_ephemeral_cache_only=use_ephemeral_cache_only
        )
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config
        )
        if cache_result:
            # print("Hit cache")
            assert type(cache_result) in [
                dict,
                BaseLMResponse,
            ], f"Expected dict or BaseLMResponse, got {type(cache_result)}"
            return (
                cache_result["response"] if type(cache_result) == dict else cache_result
            )
        if model in OPENAI_REASONING_MODELS:
            output = await self.async_client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                temperature=lm_config.get(
                    "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
                ),
                response_format=response_model,
                reasoning_effort=reasoning_effort,
            )
        else:
            output = await self.async_client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                response_format=response_model,
            )
        # "Output", output)
        api_result = response_model(**json.loads(output.choices[0].message.content))
        lm_response = BaseLMResponse(
            raw_response="",
            structured_output=api_result,
            tool_calls=None,
        )
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config, output=lm_response
        )
        return lm_response

    def _hit_api_sync_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        tools: Optional[List[BaseTool]] = None,
        reasoning_effort: str = "high",
    ) -> str:
        if tools:
            raise ValueError("Tools are not supported for sync structured output")
        lm_config = {"temperature": temperature, "response_model": response_model, "reasoning_effort": reasoning_effort}
        used_cache_handler = get_cache_handler(
            use_ephemeral_cache_only=use_ephemeral_cache_only
        )
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config
        )
        if cache_result:
            assert type(cache_result) in [
                dict,
                BaseLMResponse,
            ], f"Expected dict or BaseLMResponse, got {type(cache_result)}"
            return (
                cache_result["response"] if type(cache_result) == dict else cache_result
            )
        if model in OPENAI_REASONING_MODELS:
            output = self.sync_client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                temperature=lm_config.get(
                    "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
                ),
                response_format=response_model,
                reasoning_effort=reasoning_effort,
            )
        else:
            output = self.sync_client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                response_format=response_model,
            )
        api_result = response_model(**json.loads(output.choices[0].message.content))

        lm_response = BaseLMResponse(
            raw_response="",
            structured_output=api_result,
            tool_calls=None,
        )
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response
        )
        return lm_response


class OpenAIPrivate(OpenAIStandard):
    def __init__(self, synth_logging: bool = True):
        if synth_logging:
            # print("Using synth logging - OpenAIPrivate")
            from synth_sdk import AsyncOpenAI, OpenAI
        else:
            # print("Not using synth logging - OpenAIPrivate")
            from openai import AsyncOpenAI, OpenAI

        self.sync_client = OpenAI()
        self.async_client = AsyncOpenAI()


if __name__ == "__main__":
    client = OpenAIStructuredOutputClient(
        sync_client=openai.OpenAI(),
        async_client=openai.AsyncOpenAI(),
        used_for_structured_outputs=True,
        exceptions_to_retry=[],
    )

    class TestModel(BaseModel):
        name: str

    sync_model_response = client._hit_api_sync_structured_output(
        model="gpt-4o-mini-2024-07-18",
        messages=[{"role": "user", "content": " What is the capital of the moon?"}],
        response_model=TestModel,
        temperature=0.0,
    )
    # print(sync_model_response)



import json
from typing import Any, Dict, List, Optional, Tuple, Type

import anthropic
import pydantic
from pydantic import BaseModel

from synth_ai.zyk.lms.caching.initialize import (
    get_cache_handler,
)
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.constants import SPECIAL_BASE_TEMPS, CLAUDE_REASONING_MODELS, SONNET_37_BUDGETS
from synth_ai.zyk.lms.vendors.core.openai_api import OpenAIStructuredOutputClient

ANTHROPIC_EXCEPTIONS_TO_RETRY: Tuple[Type[Exception], ...] = (anthropic.APIError,)


class AnthropicAPI(VendorBase):
    used_for_structured_outputs: bool = True
    exceptions_to_retry: Tuple = ANTHROPIC_EXCEPTIONS_TO_RETRY
    sync_client: Any
    async_client: Any

    def __init__(
        self,
        exceptions_to_retry: Tuple[
            Type[Exception], ...
        ] = ANTHROPIC_EXCEPTIONS_TO_RETRY,
        used_for_structured_outputs: bool = False,
        reasoning_effort: str = "high",
    ):
        self.sync_client = anthropic.Anthropic()
        self.async_client = anthropic.AsyncAnthropic()
        self.used_for_structured_outputs = used_for_structured_outputs
        self.exceptions_to_retry = exceptions_to_retry
        self._openai_fallback = None
        self.reasoning_effort = reasoning_effort

    # @backoff.on_exception(
    #     backoff.expo,
    #     exceptions_to_retry,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_giveup=lambda e: print(e),
    # )
    async def _hit_api_async(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
        **vendor_params: Dict[str, Any],
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        # Common API parameters
        api_params = {
            "system": messages[0]["content"],
            "messages": messages[1:],
            "model": model,
            "max_tokens": lm_config.get("max_tokens", 4096),
            "temperature": lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            ),
        }

        # Add tools if provided
        if tools:
            api_params["tools"] = [tool.to_anthropic_tool() for tool in tools]

        # Only try to add thinking if supported by the SDK
        try:
            import inspect

            create_sig = inspect.signature(self.async_client.messages.create)
            if "thinking" in create_sig.parameters and model in CLAUDE_REASONING_MODELS:
                if reasoning_effort in ["high", "medium"]:
                    budget = SONNET_37_BUDGETS[reasoning_effort]    
                    api_params["thinking"] = {
                        "type": "enabled",
                        "budget_tokens": budget,
                    }
                    api_params["max_tokens"] = budget+4096
                    api_params["temperature"] = 1
        except (ImportError, AttributeError, TypeError):
            pass

        # Make the API call
        response = await self.async_client.messages.create(**api_params)

        # Extract text content and tool calls
        raw_response = ""
        tool_calls = []

        for content in response.content:
            if content.type == "text":
                raw_response += content.text
            elif content.type == "tool_use":
                tool_calls.append(
                    {
                        "id": content.id,
                        "type": "function",
                        "function": {
                            "name": content.name,
                            "arguments": json.dumps(content.input),
                        },
                    }
                )

        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls if tool_calls else None,
        )

        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    # @backoff.on_exception(
    #     backoff.expo,
    #     exceptions_to_retry,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_giveup=lambda e: print(e),
    # )
    def _hit_api_sync(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
        **vendor_params: Dict[str, Any],
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        used_cache_handler = get_cache_handler(
            use_ephemeral_cache_only=use_ephemeral_cache_only
        )
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        # Common API parameters
        api_params = {
            "system": messages[0]["content"],
            "messages": messages[1:],
            "model": model,
            "max_tokens": lm_config.get("max_tokens", 4096),
            "temperature": lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            ),
        }

        # Add tools if provided
        if tools:
            api_params["tools"] = [tool.to_anthropic_tool() for tool in tools]

        # Only try to add thinking if supported by the SDK
        try:
            import inspect

            create_sig = inspect.signature(self.sync_client.messages.create)
            if "thinking" in create_sig.parameters and model in CLAUDE_REASONING_MODELS:
                api_params["temperature"] = 1
                if reasoning_effort in ["high", "medium"]:
                    budgets = SONNET_37_BUDGETS
                    budget = budgets[reasoning_effort]
                    api_params["thinking"] = {
                        "type": "enabled",
                        "budget_tokens": budget,
                    }
                    api_params["max_tokens"] = budget+4096
                    api_params["temperature"] = 1
        except (ImportError, AttributeError, TypeError):
            pass

        # Make the API call
        response = self.sync_client.messages.create(**api_params)

        # Extract text content and tool calls
        raw_response = ""
        tool_calls = []

        for content in response.content:
            if content.type == "text":
                raw_response += content.text
            elif content.type == "tool_use":
                tool_calls.append(
                    {
                        "id": content.id,
                        "type": "function",
                        "function": {
                            "name": content.name,
                            "arguments": json.dumps(content.input),
                        },
                    }
                )

        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls if tool_calls else None,
        )

        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    async def _hit_api_async_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        **vendor_params: Dict[str, Any],
    ) -> BaseLMResponse:
        try:
            # First try with Anthropic
            reasoning_effort = vendor_params.get("reasoning_effort", reasoning_effort)
            if model in CLAUDE_REASONING_MODELS:

                #if reasoning_effort in ["high", "medium"]:
                budgets = SONNET_37_BUDGETS
                budget = budgets[reasoning_effort]
                max_tokens = budget+4096
                temperature = 1
                
                response = await self.async_client.messages.create(
                    system=messages[0]["content"],
                    messages=messages[1:],
                    model=model,
                    max_tokens=max_tokens,
                    thinking={"type": "enabled", "budget_tokens": budget},
                    temperature=temperature,
                )
            else:
                response = await self.async_client.messages.create(
                    system=messages[0]["content"],
                    messages=messages[1:],
                    model=model,
                    max_tokens=max_tokens,
                    temperature=temperature,
                )
            result = response.content[0].text
            parsed = json.loads(result)
            lm_response = BaseLMResponse(
                raw_response="",
                structured_output=response_model(**parsed),
                tool_calls=None,
            )
            return lm_response
        except (json.JSONDecodeError, pydantic.ValidationError):
            # If Anthropic fails, fallback to OpenAI
            if self._openai_fallback is None:
                self._openai_fallback = OpenAIStructuredOutputClient()
            return await self._openai_fallback._hit_api_async_structured_output(
                model="gpt-4o",  # Fallback to GPT-4
                messages=messages,
                response_model=response_model,
                temperature=temperature,
                use_ephemeral_cache_only=use_ephemeral_cache_only,
            )

    def _hit_api_sync_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        **vendor_params: Dict[str, Any],
    ) -> BaseLMResponse:
        try:
            # First try with Anthropic
            reasoning_effort = vendor_params.get("reasoning_effort", reasoning_effort)
            import time

            if model in CLAUDE_REASONING_MODELS:
                if reasoning_effort in ["high", "medium"]:
                    budgets = SONNET_37_BUDGETS
                    budget = budgets[reasoning_effort]
                    max_tokens = budget+4096
                    temperature = 1
                response = self.sync_client.messages.create(
                    system=messages[0]["content"],
                    messages=messages[1:],
                    model=model,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    thinking={"type": "enabled", "budget_tokens": budget},
                )
            else:
                response = self.sync_client.messages.create(
                    system=messages[0]["content"],
                    messages=messages[1:],
                    model=model,
                    max_tokens=max_tokens,
                    temperature=temperature,
                )
            # print("Time taken for API call", time.time() - t)
            result = response.content[0].text
            # Try to parse the result as JSON
            parsed = json.loads(result)
            lm_response = BaseLMResponse(
                raw_response="",
                structured_output=response_model(**parsed),
                tool_calls=None,
            )
            return lm_response
        except (json.JSONDecodeError, pydantic.ValidationError):
            # If Anthropic fails, fallback to OpenAI
            print("WARNING - Falling back to OpenAI - THIS IS SLOW")
            if self._openai_fallback is None:
                self._openai_fallback = OpenAIStructuredOutputClient()
            return self._openai_fallback._hit_api_sync_structured_output(
                model="gpt-4o",  # Fallback to GPT-4
                messages=messages,
                response_model=response_model,
                temperature=temperature,
                use_ephemeral_cache_only=use_ephemeral_cache_only,
            )

    async def _process_call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        api_call_method,
        temperature: float = 0.0,
        use_ephemeral_cache_only: bool = False,
        vendor_params: Dict[str, Any] = None,
    ) -> BaseModel:
        vendor_params = vendor_params or {}
        # Each vendor can filter parameters they support
        return await api_call_method(
            messages=messages,
            model=model,
            temperature=temperature,
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            **vendor_params,  # Pass all vendor-specific params
        )



import json
import os
from typing import Any, Dict, List, Optional, Tuple, Type

import pydantic
from mistralai import Mistral  # use Mistral as both sync and async client
from pydantic import BaseModel

from synth_ai.zyk.lms.caching.initialize import get_cache_handler
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.constants import SPECIAL_BASE_TEMPS
from synth_ai.zyk.lms.vendors.core.openai_api import OpenAIStructuredOutputClient

# Since the mistralai package doesn't expose an exceptions module,
# we fallback to catching all Exceptions for retry.
MISTRAL_EXCEPTIONS_TO_RETRY: Tuple[Type[Exception], ...] = (Exception,)


class MistralAPI(VendorBase):
    used_for_structured_outputs: bool = True
    exceptions_to_retry: Tuple = MISTRAL_EXCEPTIONS_TO_RETRY
    _openai_fallback: Any

    def __init__(
        self,
        exceptions_to_retry: Tuple[Type[Exception], ...] = MISTRAL_EXCEPTIONS_TO_RETRY,
        used_for_structured_outputs: bool = False,
    ):
        self.used_for_structured_outputs = used_for_structured_outputs
        self.exceptions_to_retry = exceptions_to_retry
        self._openai_fallback = None

    # @backoff.on_exception(
    #     backoff.expo,
    #     MISTRAL_EXCEPTIONS_TO_RETRY,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_giveup=lambda e: print(e),
    # )
    async def _hit_api_async(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        response_model: Optional[BaseModel] = None,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        assert not (response_model and tools), "Cannot provide both response_model and tools"
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            assert type(cache_result) in [
                BaseLMResponse,
                str,
            ], f"Expected BaseLMResponse or str, got {type(cache_result)}"
            return (
                cache_result
                if type(cache_result) == BaseLMResponse
                else BaseLMResponse(
                    raw_response=cache_result, structured_output=None, tool_calls=None
                )
            )

        mistral_messages = [
            {"role": msg["role"], "content": msg["content"]} for msg in messages
        ]
        functions = [tool.to_mistral_tool() for tool in tools] if tools else None
        params = {
            "model": model,
            "messages": mistral_messages,
            "max_tokens": lm_config.get("max_tokens", 4096),
            "temperature": lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            ),
            "stream": False,
            "tool_choice": "auto" if functions else None,
            
        }
        if response_model:
            params["response_format"] = response_model
        elif tools:
            params["tools"] = functions

        async with Mistral(api_key=os.getenv("MISTRAL_API_KEY", "")) as client:
            response = await client.chat.complete_async(**params)

        message = response.choices[0].message
        try:
            raw_response = message.content
        except AttributeError:
            raw_response = ""

        tool_calls = []
        try:
            if message.tool_calls:
                tool_calls = [
                    {
                        "id": call.id,
                        "type": "function",
                        "function": {
                            "name": call.function.name,
                            "arguments": call.function.arguments,
                        },
                    }
                    for call in message.tool_calls
                ]
        except AttributeError:
            pass

        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls if tool_calls else None,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    # @backoff.on_exception(
    #     backoff.expo,
    #     MISTRAL_EXCEPTIONS_TO_RETRY,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_giveup=lambda e: print(e),
    # )
    def _hit_api_sync(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        response_model: Optional[BaseModel] = None,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        assert not (response_model and tools), "Cannot provide both response_model and tools"
       
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            assert type(cache_result) in [
                BaseLMResponse,
                str,
            ], f"Expected BaseLMResponse or str, got {type(cache_result)}"
            return (
                cache_result
                if type(cache_result) == BaseLMResponse
                else BaseLMResponse(
                    raw_response=cache_result, structured_output=None, tool_calls=None
                )
            )

        mistral_messages = [
            {"role": msg["role"], "content": msg["content"]} for msg in messages
        ]
        functions = [tool.to_mistral_tool() for tool in tools] if tools else None
        
        params = {
            "model": model,
            "messages": mistral_messages,
            "max_tokens": lm_config.get("max_tokens", 4096),
            "temperature": lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            ),
            "stream": False,
            "tool_choice": "auto" if functions else None,
            #"tools": functions,
        }
        if response_model:
            params["response_format"] = response_model
        elif tools:
            params["tools"] = functions

        with Mistral(api_key=os.getenv("MISTRAL_API_KEY", "")) as client:
            response = client.chat.complete(**params)

        message = response.choices[0].message
        try:
            raw_response = message.content
        except AttributeError:
            raw_response = ""

        tool_calls = []
        try:
            if message.tool_calls:
                tool_calls = [
                    {
                        "id": call.id,
                        "type": "function",
                        "function": {
                            "name": call.function.name,
                            "arguments": call.function.arguments,
                        },
                    }
                    for call in message.tool_calls
                ]
        except AttributeError:
            pass

        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls if tool_calls else None,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    async def _hit_api_async_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
    ) -> BaseLMResponse:
        try:
            mistral_messages = [
                {"role": msg["role"], "content": msg["content"]} for msg in messages
            ]
            async with Mistral(api_key=os.getenv("MISTRAL_API_KEY", "")) as client:
                response = await client.chat.complete_async(
                    model=model,
                    messages=mistral_messages,
                    max_tokens=4096,
                    temperature=temperature,
                    stream=False,
                )
            result = response.choices[0].message.content
            parsed = json.loads(result)
            lm_response = BaseLMResponse(
                raw_response="",
                structured_output=response_model(**parsed),
                tool_calls=None,
            )
            return lm_response
        except (json.JSONDecodeError, pydantic.ValidationError):
            if self._openai_fallback is None:
                self._openai_fallback = OpenAIStructuredOutputClient()
            return await self._openai_fallback._hit_api_async_structured_output(
                model="gpt-4o",
                messages=messages,
                response_model=response_model,
                temperature=temperature,
                use_ephemeral_cache_only=use_ephemeral_cache_only,
            )

    def _hit_api_sync_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
    ) -> BaseLMResponse:
        try:
            mistral_messages = [
                {"role": msg["role"], "content": msg["content"]} for msg in messages
            ]
            with Mistral(api_key=os.getenv("MISTRAL_API_KEY", "")) as client:
                response = client.chat.complete(
                    model=model,
                    messages=mistral_messages,
                    max_tokens=4096,
                    temperature=temperature,
                    stream=False,
                )
            result = response.choices[0].message.content
            parsed = json.loads(result)
            lm_response = BaseLMResponse(
                raw_response="",
                structured_output=response_model(**parsed),
                tool_calls=None,
            )
            return lm_response
        except (json.JSONDecodeError, pydantic.ValidationError):
            print("WARNING - Falling back to OpenAI - THIS IS SLOW")
            if self._openai_fallback is None:
                self._openai_fallback = OpenAIStructuredOutputClient()
            return self._openai_fallback._hit_api_sync_structured_output(
                model="gpt-4o",
                messages=messages,
                response_model=response_model,
                temperature=temperature,
                use_ephemeral_cache_only=use_ephemeral_cache_only,
            )


if __name__ == "__main__":
    import asyncio

    from pydantic import BaseModel

    class TestModel(BaseModel):
        name: str

    client = MistralAPI(used_for_structured_outputs=True, exceptions_to_retry=[])
    import time

    t = time.time()

    async def run_async():
        response = await client._hit_api_async_structured_output(
            model="mistral-large-latest",
            messages=[{"role": "user", "content": "What is the capital of the moon?"}],
            response_model=TestModel,
            temperature=0.0,
        )
        print(response)
        return response

    response = asyncio.run(run_async())
    t2 = time.time()
    print(f"Got {len(response.name)} chars in {t2-t} seconds")



from typing import Any, Dict, List, Optional, Union

import groq
import openai
import pydantic_core
from pydantic import BaseModel

from synth_ai.zyk.lms.caching.initialize import (
    get_cache_handler,
)
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.constants import SPECIAL_BASE_TEMPS

DEFAULT_EXCEPTIONS_TO_RETRY = (
    pydantic_core._pydantic_core.ValidationError,
    openai.APIConnectionError,
    openai.APITimeoutError,
    groq.InternalServerError,
    groq.APITimeoutError,
    groq.APIConnectionError,
)


def special_orion_transform(
    model: str, messages: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    if "o1-" in model:
        messages = [
            {
                "role": "user",
                "content": f"<instructions>{messages[0]['content']}</instructions><information>{messages[1]}</information>",
            }
        ]
    return messages


def on_backoff_handler_async(details):
    # Print every 5th retry attempt, excluding the first retry
    if details["tries"] > 1 and (details["tries"] - 1) % 5 == 0:
        print(f"Retrying async API call (attempt {details['tries'] - 1})")


def on_backoff_handler_sync(details):
    # Print every 5th retry attempt, excluding the first retry
    if details["tries"] > 1 and (details["tries"] - 1) % 5 == 0:
        print(f"Retrying sync API call (attempt {details['tries'] - 1})")


class OpenAIStandard(VendorBase):
    used_for_structured_outputs: bool = True
    exceptions_to_retry: List = DEFAULT_EXCEPTIONS_TO_RETRY
    sync_client: Any
    async_client: Any

    def __init__(
        self,
        sync_client: Any,
        async_client: Any,
        exceptions_to_retry: List[Exception] = DEFAULT_EXCEPTIONS_TO_RETRY,
        used_for_structured_outputs: bool = False,
    ):
        self.sync_client = sync_client
        self.async_client = async_client
        self.used_for_structured_outputs = used_for_structured_outputs
        self.exceptions_to_retry = exceptions_to_retry

    # @backoff.on_exception(
    #     backoff.expo,
    #     exceptions_to_retry,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_backoff=on_backoff_handler_async,
    #     on_giveup=lambda e: print(e),
    # )
    async def _hit_api_async(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        messages = special_orion_transform(model, messages)
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        # Common API call params
        api_params = {
            "model": model,
            "messages": messages,
        }

        # Add tools if provided
        if tools and all(isinstance(tool, BaseTool) for tool in tools):
            api_params["tools"] = [tool.to_openai_tool() for tool in tools]
        elif tools:
            api_params["tools"] = tools

        # Only add temperature for non o1/o3 models
        if not any(prefix in model for prefix in ["o1-", "o3-"]):
            api_params["temperature"] = lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            )

        # Add reasoning_effort only for o3-mini
        if model in ["o3-mini"]:
            print("Reasoning effort:", reasoning_effort)
            api_params["reasoning_effort"] = reasoning_effort

        output = await self.async_client.chat.completions.create(**api_params)
        message = output.choices[0].message

        # Convert tool calls to dict format
        tool_calls = None
        if message.tool_calls:
            tool_calls = [
                {
                    "id": tc.id,
                    "type": tc.type,
                    "function": {
                        "name": tc.function.name,
                        "arguments": tc.function.arguments,
                    },
                }
                for tc in message.tool_calls
            ]

        lm_response = BaseLMResponse(
            raw_response=message.content or "",  # Use empty string if no content
            structured_output=None,
            tool_calls=tool_calls,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    # @backoff.on_exception(
    #     backoff.expo,
    #     exceptions_to_retry,
    #     max_tries=BACKOFF_TOLERANCE,
    #     on_backoff=on_backoff_handler_sync,
    #     on_giveup=lambda e: print(e),
    # )
    def _hit_api_sync(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        messages = special_orion_transform(model, messages)
        used_cache_handler = get_cache_handler(
            use_ephemeral_cache_only=use_ephemeral_cache_only
        )
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        # Common API call params
        api_params = {
            "model": model,
            "messages": messages,
        }

        # Add tools if provided
        if tools and all(isinstance(tool, BaseTool) for tool in tools):
            api_params["tools"] = [tool.to_openai_tool() for tool in tools]
        elif tools:
            api_params["tools"] = tools

        # Only add temperature for non o1/o3 models
        if not any(prefix in model for prefix in ["o1-", "o3-"]):
            api_params["temperature"] = lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            )

        # Add reasoning_effort only for o3-mini
        if model in ["o3-mini"]:
            api_params["reasoning_effort"] = reasoning_effort

        output = self.sync_client.chat.completions.create(**api_params)
        message = output.choices[0].message

        # Convert tool calls to dict format
        tool_calls = None
        if message.tool_calls:
            tool_calls = [
                {
                    "id": tc.id,
                    "type": tc.type,
                    "function": {
                        "name": tc.function.name,
                        "arguments": tc.function.arguments,
                    },
                }
                for tc in message.tool_calls
            ]

        lm_response = BaseLMResponse(
            raw_response=message.content or "",  # Use empty string if no content
            structured_output=None,
            tool_calls=tool_calls,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    async def _hit_api_async_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        lm_config = {"temperature": temperature, "response_model": response_model, "reasoning_effort": reasoning_effort}
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        cache_result: Union[BaseLMResponse, None] = (
            used_cache_handler.hit_managed_cache(
                model, messages, lm_config=lm_config, tools=tools
            )
        )
        if cache_result is not None:
            return cache_result

        # Common API call params
        api_params = {
            "model": model,
            "messages": messages,
        }

        # Add tools if provided
        if tools and all(isinstance(tool, BaseTool) for tool in tools):
            api_params["tools"] = [tool.to_openai_tool() for tool in tools]
        elif tools:
            api_params["tools"] = tools

        # Only add temperature for non o1/o3 models
        if not any(prefix in model for prefix in ["o1-", "o3-"]):
            api_params["temperature"] = lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            )

        # Add reasoning_effort only for o3-mini
        if model in ["o3-mini"]:
            api_params["reasoning_effort"] = reasoning_effort

        output = await self.async_client.chat.completions.create(**api_params)

        structured_output_api_result = response_model(
            **output.choices[0].message.content
        )
        tool_calls = output.choices[0].message.tool_calls
        lm_response = BaseLMResponse(
            raw_response=output.choices[0].message.content,
            structured_output=structured_output_api_result,
            tool_calls=tool_calls,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    def _hit_api_sync_structured_output(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        temperature: float,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        lm_config = {"temperature": temperature, "response_model": response_model, "reasoning_effort": reasoning_effort}
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        cache_result: Union[BaseLMResponse, None] = (
            used_cache_handler.hit_managed_cache(
                model, messages, lm_config=lm_config, tools=tools
            )
        )
        if cache_result is not None:
            return cache_result

        # Common API call params
        api_params = {
            "model": model,
            "messages": messages,
        }

        # Add tools if provided
        if tools and all(isinstance(tool, BaseTool) for tool in tools):
            api_params["tools"] = [tool.to_openai_tool() for tool in tools]
        elif tools:
            api_params["tools"] = tools

        # Only add temperature for non o1/o3 models
        if not any(prefix in model for prefix in ["o1-", "o3-"]):
            api_params["temperature"] = lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0)
            )

        # Add reasoning_effort only for o3-mini
        if model in ["o3-mini"]:
            api_params["reasoning_effort"] = reasoning_effort

        output = self.sync_client.chat.completions.create(**api_params)

        structured_output_api_result = response_model(
            **output.choices[0].message.content
        )
        tool_calls = output.choices[0].message.tool_calls
        lm_response = BaseLMResponse(
            raw_response=output.choices[0].message.content,
            structured_output=structured_output_api_result,
            tool_calls=tool_calls,
        )
        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response



import re
from typing import Any, List, Pattern

from synth_ai.zyk.lms.core.all import (
    AnthropicClient,
    DeepSeekClient,
    GeminiClient,
    GroqClient,
    MistralClient,
    # OpenAIClient,
    OpenAIStructuredOutputClient,
    TogetherClient,
    CustomEndpointClient,
)

openai_naming_regexes: List[Pattern] = [
    re.compile(r"^(ft:)?(o[1,3,4](-.*)?|gpt-.*)$"),
]
openai_formatting_model_regexes: List[Pattern] = [
    re.compile(r"^(ft:)?gpt-4o(-.*)?$"),
]
anthropic_naming_regexes: List[Pattern] = [
    re.compile(r"^claude-.*$"),
]
gemini_naming_regexes: List[Pattern] = [
    re.compile(r"^gemini-.*$"),
    re.compile(r"^gemma[2-9].*$"),
]
deepseek_naming_regexes: List[Pattern] = [
    re.compile(r"^deepseek-.*$"),
]
together_naming_regexes: List[Pattern] = [
    re.compile(r"^.*\/.*$"),
]

groq_naming_regexes: List[Pattern] = [
    re.compile(r"^llama-3.3-70b-versatile$"),
    re.compile(r"^llama-3.1-8b-instant$"),
    re.compile(r"^qwen-2.5-32b$"),
    re.compile(r"^deepseek-r1-distill-qwen-32b$"),
    re.compile(r"^deepseek-r1-distill-llama-70b-specdec$"),
    re.compile(r"^deepseek-r1-distill-llama-70b$"),
    re.compile(r"^llama-3.3-70b-specdec$"),
    re.compile(r"^llama-3.2-1b-preview$"),
    re.compile(r"^llama-3.2-3b-preview$"),
    re.compile(r"^llama-3.2-11b-vision-preview$"),
    re.compile(r"^llama-3.2-90b-vision-preview$"),
    re.compile(r"^meta-llama/llama-4-scout-17b-16e-instruct$"),
    re.compile(r"^meta-llama/llama-4-maverick-17b-128e-instruct$"),
    re.compile(r"^qwen/qwen3-32b$"),
]

mistral_naming_regexes: List[Pattern] = [
    re.compile(r"^mistral-.*$"),
]

# Custom endpoint patterns - check these before generic patterns
custom_endpoint_naming_regexes: List[Pattern] = [
    # Modal endpoints: org--app.modal.run
    re.compile(r"^[a-zA-Z0-9\-]+--[a-zA-Z0-9\-]+\.modal\.run$"),
    # Generic domain patterns for custom endpoints
    re.compile(r"^[a-zA-Z0-9\-]+\.[a-zA-Z0-9\-]+\.[a-zA-Z]+$"),  # domain.tld
    re.compile(r"^[a-zA-Z0-9\-]+\.[a-zA-Z0-9\-]+\.[a-zA-Z]+\/[a-zA-Z0-9\-\/]+$"),  # domain.tld/path
]


def get_client(
    model_name: str,
    with_formatting: bool = False,
    synth_logging: bool = True,
) -> Any:
    # print("With formatting", with_formatting)
    if any(regex.match(model_name) for regex in openai_naming_regexes):
        # print("Returning OpenAIStructuredOutputClient")
        return OpenAIStructuredOutputClient(
            synth_logging=synth_logging,
        )
    elif any(regex.match(model_name) for regex in anthropic_naming_regexes):
        if with_formatting:
            client = AnthropicClient()
            client._hit_api_async_structured_output = OpenAIStructuredOutputClient(
                synth_logging=synth_logging
            )._hit_api_async
            return client
        else:
            return AnthropicClient()
    elif any(regex.match(model_name) for regex in gemini_naming_regexes):
        return GeminiClient()
    elif any(regex.match(model_name) for regex in deepseek_naming_regexes):
        return DeepSeekClient()
    elif any(regex.match(model_name) for regex in groq_naming_regexes):
        return GroqClient()
    elif any(regex.match(model_name) for regex in mistral_naming_regexes):
        return MistralClient()
    elif any(regex.match(model_name) for regex in custom_endpoint_naming_regexes):
        # Custom endpoints are passed as the endpoint URL
        return CustomEndpointClient(endpoint_url=model_name)
    elif any(regex.match(model_name) for regex in together_naming_regexes):
        return TogetherClient()
    else:
        raise ValueError(f"Invalid model name: {model_name}")



"""
Synth AI - Software for aiding the best and multiplying the will.
"""

from synth_ai.zyk import LM  # Assuming LM is in zyk.py in the same directory

__version__ = "0.1.5"
__all__ = ["LM"]  # Explicitly define public API



from openai import OpenAI, AsyncOpenAI
from synth_ai.zyk.lms.vendors.openai_standard import OpenAIStandard


class OllamaAPI(OpenAIStandard):
    def __init__(self):
        self.sync_client = OpenAI(
            base_url="http://localhost:11434/v1",
            api_key="ollama",  # required, but unused
        )
        self.async_client = AsyncOpenAI(
            base_url="http://localhost:11434/v1",
            api_key="ollama",  # required, but unused
        )


import unittest
from typing import List, Union

from pydantic import BaseModel

from synth_ai.zyk.lms.core.main import LM


class UnitTestDict(BaseModel):
    test_description: str
    input_names: List[str]
    input_types: List[str]
    stringified_input_values: List[str]
    assertion_condition: str
    assertion_type: str  # Consider using Literal for specific assertion types


class ActionArgument(BaseModel):
    key: str
    value: Union[str, int, float, bool, UnitTestDict]


class ReAct(BaseModel):
    reasoning: str
    action_name: str
    action_args: List[ActionArgument]  # Dict#[str, Dict]


system = """
<System Message> 
# Premise
You are a software engineer
Here is some information about this setting
<Setting Information>
You are working to solve a computer science problem. You will need to submit a solution to the problem, which will be tested against a suite of hidden unit tests.
</Setting Information>
<Actions Available>
<edit_submission>
<action_context>
Edit the submission code. Use this when you want to make changes to the current solution.
</action_context>
<action_arg_spec>
{'first_line': <class 'int'>, 'last_line': <class 'int'>, 'new_code': <class 'str'>}
</action_arg_spec>
<action_description>
Edit the submission code
</action_description>

</edit_submission>
<add_submission>
<action_context>
Add the submission code. Use this when you want to start from scratch with a new solution.
</action_context>
<action_arg_spec>
{'submission': <class 'str'>}
</action_arg_spec>
<action_description>
Add the submission code
</action_description>

</add_submission>
<add_unit_test>
<action_context>
Add a unit test. The unit test information you submit must be in the format of a BCBUnitTest: 

class BCBUnitTest(BaseModel):
    test_description: str
    input_names: List[str]
    input_types: List[str]
    input_values: List[Any]
    assertion_condition: str
    assertion_type: Literal["assertTrue", "assertRaises"] = "assertTrue"


 It will be parsed via BCBUnitTest(**unit_test_dict)



# Some various notes:
1. If an input should be of a type defined by a specific package, add the package name/alias to the type. E.g. "np.ndarray" or "pd.DataFrame". You still should fully define the value for the input_value field e.g. "pd.DataFrame({'a': [1, 2, 3]})"

2. Unit tests will be compiled from the BCBUnitTest class as follows:
    A. For AssertTrue type tests, the test will be compiled as follows:
    ```python
    def test_case(self):
        # {{self.test_description}}

        {{defs}}
        result = {{function_name}}(**{{{{args}}}}})
        self.{{self.assertion_type}}({{self.assertion_condition}})
    ```
    B. For AssertRaises type tests, the test will be compiled as follows:

    ```python
    def test_case(self):
        # {{self.test_description}}
        {{defs}}
        with self.{{self.assertion_type}}({{self.assertion_condition}}):
            {{function_name}}(**{{{{args}}}}})
    ```

    Provide information accordingly.

</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>, 'unit_test_dict': typing.Dict}
</action_arg_spec>
<action_description>
Add a unit test
</action_description>

</add_unit_test>
<remove_unit_test>
<action_context>
Remove a unit test
</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>}
</action_arg_spec>
<action_description>
Remove a unit test
</action_description>

</remove_unit_test>
<test_submission>
<action_context>
Test the submission
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Test the submission
</action_description>

</test_submission>
<submit_solution>
<action_context>
Submit the solution
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Submit the solution
</action_description>

</submit_solution>

</Actions Available>
You'll be given your past actions/thoughts, along with recent raw observations from the environment
The environment one step in the past is your current environment.

# Objective
Please complete the problem by drafting a solution, creating unit tests, improving the solution, and submitting the solution.

# Constraints
You will be given a code_prompt_for_answer, which contains imports and the function signature. Your solution must comprise code that can be appended to code_prompt_for_answer and run as a single script.
 
"""

user = """
<User Message> 
# Recent Actions / Thoughts

# Recent Observations
<1 environment step(s) in the past>{'action_result': None, 'environment_state': {'question': 'import pandas as pd\nimport numpy as np\n\n# Constants\nCOLUMNS = [\'column1\', \'column2\', \'column3\', \'column4\', \'column5\']\n\ndef task_func(df, dct):\n    '''\n    Replace certain values in a DataFrame with a dictionary mapping and calculate the Pearson correlation coefficient between each pair of columns.\n\n    Parameters:\n    df (DataFrame): The input DataFrame, containing numeric or categorical data.\n    dct (dict): A dictionary for replacing values in df, where keys are existing values and values are new values.\n\n    Returns:\n    DataFrame: A DataFrame with the correlation coefficients between each pair of columns. The format of the DataFrame is a square matrix with column and index labels matching the columns of the input DataFrame.\n    \n    Requirements:\n    - pandas\n    - numpy\n    \n    Note:\n    - This function operates on DataFrames containing numeric or categorical data that can be replaced with numeric values, as correlation calculations require numeric data.\n    - This function using pearson method to calculate the correlation matrix.\n    \n    Raises:\n    - This function will raise a ValueError is input df is not a DataFrame.\n        \n    Example:\n    >>> df = pd.DataFrame({\'A\': [1, 2, 3], \'B\': [4, 5, 6]})\n    >>> dct = {1: 10, 2: 20, 3: 30, 4: 40, 5: 50, 6: 60}\n    >>> correlation_matrix = task_func(df, dct)\n    >>> correlation_matrix.shape == (2, 2)\n    True\n    >>> np.allclose(correlation_matrix, np.array([[1.0, 1.0], [1.0, 1.0]]))\n    True\n    '''\n', 'code_prompt_for_answer': "import pandas as pd\nimport numpy as np\n# Constants\nCOLUMNS = ['column1', 'column2', 'column3', 'column4', 'column5']\ndef task_func(df, dct):\n", 'unit_tests_you_have_written': {}, 'current_solution': ''}}</1 environment step(s) in the past>

Your next actions / thought:
"""


###

hard_system = """
# Premise
You are a software engineer
Here is some information about this setting
<Setting Information>
You are working to solve a computer science problem. You will need to submit a solution to the problem, which will be tested against a suite of hidden unit tests.
</Setting Information>
<Actions Available>
<edit_submission>
<action_context>
Edit the submission code. Use this when you want to make changes to the current solution.
</action_context>
<action_arg_spec>
{'first_line': <class 'int'>, 'last_line': <class 'int'>, 'new_code': <class 'str'>}
</action_arg_spec>
<action_description>
Edit the submission code
</action_description>

</edit_submission>
<add_submission>
<action_context>
Add the submission code. Use this when you want to start from scratch with a new solution.
</action_context>
<action_arg_spec>
{'submission': <class 'str'>}
</action_arg_spec>
<action_description>
Add the submission code
</action_description>

</add_submission>
<add_unit_test>
<action_context>
Add a unit test. The unit test information you submit must be in the format of a BCBUnitTest: 

class BCBUnitTest(BaseModel):
    test_description: str
    input_names: List[str]
    input_types: List[str]
    input_values: List[Any]
    assertion_condition: str
    assertion_type: Literal["assertTrue", "assertRaises"] = "assertTrue"


 It will be parsed via BCBUnitTest(**unit_test_dict)



# Some various notes:
1. If an input should be of a type defined by a specific package, add the package name/alias to the type. E.g. "np.ndarray" or "pd.DataFrame". You still should fully define the value for the input_value field e.g. "pd.DataFrame({'a': [1, 2, 3]})"

2. Unit tests will be compiled from the BCBUnitTest class as follows:
    A. For AssertTrue type tests, the test will be compiled as follows:
    ```python
    def test_case(self):
        # {{self.test_description}}

        {{defs}}
        result = {{function_name}}(**{{{{args}}}}})
        self.{{self.assertion_type}}({{self.assertion_condition}})
    ```
    B. For AssertRaises type tests, the test will be compiled as follows:

    ```python
    def test_case(self):
        # {{self.test_description}}
        {{defs}}
        with self.{{self.assertion_type}}({{self.assertion_condition}}):
            {{function_name}}(**{{{{args}}}}})
    ```

    Provide information accordingly.

</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>, 'unit_test_dict': typing.Dict}
</action_arg_spec>
<action_description>
Add a unit test
</action_description>

</add_unit_test>
<remove_unit_test>
<action_context>
Remove a unit test
</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>}
</action_arg_spec>
<action_description>
Remove a unit test
</action_description>

</remove_unit_test>
<test_submission>
<action_context>
Test the submission
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Test the submission
</action_description>

</test_submission>
<submit_solution>
<action_context>
Submit the solution
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Submit the solution
</action_description>

</submit_solution>

</Actions Available>
You'll be given your past actions/thoughts, along with recent raw observations from the environment
The environment one step in the past is your current environment.

# Objective
Please complete the problem by drafting a solution, creating unit tests, improving the solution, and submitting the solution.

# Constraints
You will be given a code_prompt_for_answer, which contains imports and the function signature. Your solution must comprise code that can be appended to code_prompt_for_answer and run as a single script.
 

<User Message> 
# Recent Actions / Thoughts

# Recent Observations
<1 environment step(s) in the past>{'action_result': None, 'environment_state': {'question': 'import pandas as pd\nimport numpy as np\n\n# Constants\nCOLUMNS = [\'column1\', \'column2\', \'column3\', \'column4\', \'column5\']\n\ndef task_func(df, dct):\n   '''\n    Replace certain values in a DataFrame with a dictionary mapping and calculate the Pearson correlation coefficient between each pair of columns.\n\n    Parameters:\n    df (DataFrame): The input DataFrame, containing numeric or categorical data.\n    dct (dict): A dictionary for replacing values in df, where keys are existing values and values are new values.\n\n    Returns:\n    DataFrame: A DataFrame with the correlation coefficients between each pair of columns. The format of the DataFrame is a square matrix with column and index labels matching the columns of the input DataFrame.\n    \n    Requirements:\n    - pandas\n    - numpy\n    \n    Note:\n    - This function operates on DataFrames containing numeric or categorical data that can be replaced with numeric values, as correlation calculations require numeric data.\n    - This function using pearson method to calculate the correlation matrix.\n    \n    Raises:\n    - This function will raise a ValueError is input df is not a DataFrame.\n        \n    Example:\n    >>> df = pd.DataFrame({\'A\': [1, 2, 3], \'B\': [4, 5, 6]})\n    >>> dct = {1: 10, 2: 20, 3: 30, 4: 40, 5: 50, 6: 60}\n    >>> correlation_matrix = task_func(df, dct)\n    >>> correlation_matrix.shape == (2, 2)\n    True\n    >>> np.allclose(correlation_matrix, np.array([[1.0, 1.0], [1.0, 1.0]]))\n    True\n    '''\n', 'code_prompt_for_answer': "import pandas as pd\nimport numpy as np\n# Constants\nCOLUMNS = ['column1', 'column2', 'column3', 'column4', 'column5']\ndef task_func(df, dct):\n", 'unit_tests_you_have_written': {}, 'current_solution': ''}}</1 environment step(s) in the past>

Your next actions / thought:  

Structured output: reasoning="I need to implement the function 'task_func' that replaces values in a DataFrame based on a dictionary and calculates the Pearson correlation coefficient between the columns. I will also ensure to handle the case where the input is not a DataFrame by raising a ValueError." action_name='edit_submission' action_args=[ActionArgument(key='first_line', value=4), ActionArgument(key='last_line', value=4), ActionArgument(key='new_code', value="    if not isinstance(df, pd.DataFrame):\n        raise ValueError('Input must be a DataFrame')\n    df.replace(dct, inplace=True)\n    return df.corr(method='pearson')")]
<System Message> 
# Premise
You are a software engineer
Here is some information about this setting
<Setting Information>
You are working to solve a computer science problem. You will need to submit a solution to the problem, which will be tested against a suite of hidden unit tests.
</Setting Information>
<Actions Available>
<edit_submission>
<action_context>
Edit the submission code. Use this when you want to make changes to the current solution.
</action_context>
<action_arg_spec>
{'first_line': <class 'int'>, 'last_line': <class 'int'>, 'new_code': <class 'str'>}
</action_arg_spec>
<action_description>
Edit the submission code
</action_description>

</edit_submission>
<add_submission>
<action_context>
Add the submission code. Use this when you want to start from scratch with a new solution.
</action_context>
<action_arg_spec>
{'submission': <class 'str'>}
</action_arg_spec>
<action_description>
Add the submission code
</action_description>

</add_submission>
<add_unit_test>
<action_context>
Add a unit test. The unit test information you submit must be in the format of a BCBUnitTest: 

class BCBUnitTest(BaseModel):
    test_description: str
    input_names: List[str]
    input_types: List[str]
    input_values: List[Any]
    assertion_condition: str
    assertion_type: Literal["assertTrue", "assertRaises"] = "assertTrue"


 It will be parsed via BCBUnitTest(**unit_test_dict)



# Some various notes:
1. If an input should be of a type defined by a specific package, add the package name/alias to the type. E.g. "np.ndarray" or "pd.DataFrame". You still should fully define the value for the input_value field e.g. "pd.DataFrame({'a': [1, 2, 3]})"

2. Unit tests will be compiled from the BCBUnitTest class as follows:
    A. For AssertTrue type tests, the test will be compiled as follows:
    ```python
    def test_case(self):
        # {{self.test_description}}

        {{defs}}
        result = {{function_name}}(**{{{{args}}}}})
        self.{{self.assertion_type}}({{self.assertion_condition}})
    ```
    B. For AssertRaises type tests, the test will be compiled as follows:

    ```python
    def test_case(self):
        # {{self.test_description}}
        {{defs}}
        with self.{{self.assertion_type}}({{self.assertion_condition}}):
            {{function_name}}(**{{{{args}}}}})
    ```

    Provide information accordingly.

</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>, 'unit_test_dict': typing.Dict}
</action_arg_spec>
<action_description>
Add a unit test
</action_description>

</add_unit_test>
<remove_unit_test>
<action_context>
Remove a unit test
</action_context>
<action_arg_spec>
{'unit_test_name': <class 'str'>}
</action_arg_spec>
<action_description>
Remove a unit test
</action_description>

</remove_unit_test>
<test_submission>
<action_context>
Test the submission
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Test the submission
</action_description>

</test_submission>
<submit_solution>
<action_context>
Submit the solution
</action_context>
<action_arg_spec>
{}
</action_arg_spec>
<action_description>
Submit the solution
</action_description>

</submit_solution>

</Actions Available>
You'll be given your past actions/thoughts, along with recent raw observations from the environment
The environment one step in the past is your current environment.

# Objective
Please complete the problem by drafting a solution, creating unit tests, improving the solution, and submitting the solution.

# Constraints
You will be given a code_prompt_for_answer, which contains imports and the function signature. Your solution must comprise code that can be appended to code_prompt_for_answer and run as a single script.
"""

hard_user = """
# Recent Actions / Thoughts
<1 reasoning step(s) in the past>reasoning="I need to implement the function 'task_func' that replaces values in a DataFrame based on a dictionary and calculates the Pearson correlation coefficient between the columns. I will also ensure to handle the case where the input is not a DataFrame by raising a ValueError." action_name='edit_submission' action_args=[ActionArgument(key='first_line', value=4), ActionArgument(key='last_line', value=4), ActionArgument(key='new_code', value="    if not isinstance(df, pd.DataFrame):\n        raise ValueError('Input must be a DataFrame')\n    df.replace(dct, inplace=True)\n    return df.corr(method='pearson')")]</1 reasoning step(s) in the past>
# Recent Observations
<1 environment step(s) in the past>success=True result='Edited submission successfully'</1 environment step(s) in the past>
<2 environment step(s) in the past>{'action_result': None, 'environment_state': {'question': 'import pandas as pd\nimport numpy as np\n\n# Constants\nCOLUMNS = [\'column1\', \'column2\', \'column3\', \'column4\', \'column5\']\n\ndef task_func(df, dct):\n    '''\n    Replace certain values in a DataFrame with a dictionary mapping and calculate the Pearson correlation coefficient between each pair of columns.\n\n    Parameters:\n    df (DataFrame): The input DataFrame, containing numeric or categorical data.\n    dct (dict): A dictionary for replacing values in df, where keys are existing values and values are new values.\n\n    Returns:\n    DataFrame: A DataFrame with the correlation coefficients between each pair of columns. The format of the DataFrame is a square matrix with column and index labels matching the columns of the input DataFrame.\n    \n    Requirements:\n    - pandas\n    - numpy\n    \n    Note:\n    - This function operates on DataFrames containing numeric or categorical data that can be replaced with numeric values, as correlation calculations require numeric data.\n    - This function using pearson method to calculate the correlation matrix.\n    \n    Raises:\n    - This function will raise a ValueError is input df is not a DataFrame.\n        \n    Example:\n    >>> df = pd.DataFrame({\'A\': [1, 2, 3], \'B\': [4, 5, 6]})\n    >>> dct = {1: 10, 2: 20, 3: 30, 4: 40, 5: 50, 6: 60}\n    >>> correlation_matrix = task_func(df, dct)\n    >>> correlation_matrix.shape == (2, 2)\n    True\n    >>> np.allclose(correlation_matrix, np.array([[1.0, 1.0], [1.0, 1.0]]))\n    True\n    '''\n', 'code_prompt_for_answer': "import pandas as pd\nimport numpy as np\n# Constants\nCOLUMNS = ['column1', 'column2', 'column3', 'column4', 'column5']\ndef task_func(df, dct):\n", 'unit_tests_you_have_written': {}, 'current_solution': ''}}</2 environment step(s) in the past>

Your next actions / thought:
"""


class TestLMStructuredOutputs(unittest.TestCase):
    # ... existing code ...

    @classmethod
    def setUpClass(cls):
        # Initialize LMs for both forced_json and stringified_json modes
        cls.lm_forced_json = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.0,
            max_retries="Few",
            structured_output_mode="forced_json",
        )
        cls.lm_stringified_json = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.0,
            max_retries="Few",
            structured_output_mode="stringified_json",
        )

    def test_sync_react_response_content(self):
        system_message = system

        user_message = user

        for lm in [self.lm_forced_json, self.lm_stringified_json]:
            with self.subTest(
                mode=lm.structured_output_handler.handler.structured_output_mode
            ):
                result = lm.respond_sync(
                    system_message=system_message,
                    user_message=user_message,
                    response_model=ReAct,
                )
                self.assertIsInstance(result, ReAct)
                self.assertIsInstance(result.reasoning, str)
                self.assertIsInstance(result.action_name, str)
                self.assertIsInstance(result.action_args, list)
                for arg in result.action_args:
                    self.assertIsInstance(arg, ActionArgument)
                    self.assertIsInstance(arg.key, str)
                    # self.assertIsInstance(arg.value, str)

    def test_sync_react_response_hard_content(self):
        system_message = hard_system

        user_message = hard_user

        for lm in [self.lm_forced_json, self.lm_stringified_json]:
            with self.subTest(
                mode=lm.structured_output_handler.handler.structured_output_mode
            ):
                result = lm.respond_sync(
                    system_message=system_message,
                    user_message=user_message,
                    response_model=ReAct,
                )
                self.assertIsInstance(result, ReAct)
                self.assertIsInstance(result.reasoning, str)
                self.assertIsInstance(result.action_name, str)
                self.assertIsInstance(result.action_args, list)
                for arg in result.action_args:
                    self.assertIsInstance(arg, ActionArgument)
                    self.assertIsInstance(arg.key, str)
                    # self.assertIsInstance(arg.value, str)


# use non-trivial fallback?

if __name__ == "__main__":
    # Create an instance of the test class
    test_instance = TestLMStructuredOutputs()

    # Set up the class (this would normally be done by unittest)
    test_instance.setUpClass()

    # Run the test methods
    test_instance.test_sync_react_response_content()
    test_instance.test_sync_react_response_hard_content()

    print("All tests completed.")



import json
import logging
import os
import warnings
from typing import Any, Dict, List, Optional, Tuple, Type

from google import genai
from google.api_core.exceptions import ResourceExhausted
from google.genai import types
from synth_ai.zyk.lms.caching.initialize import get_cache_handler
from synth_ai.zyk.lms.tools.base import BaseTool
from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.constants import (
    SPECIAL_BASE_TEMPS,
    GEMINI_REASONING_MODELS,
    GEMINI_THINKING_BUDGETS,
)
from synth_ai.zyk.lms.vendors.retries import BACKOFF_TOLERANCE, backoff
import logging


ALIASES = {
    "gemini-2.5-flash": "gemini-2.5-flash-preview-04-17",
}

logger = logging.getLogger(__name__)
_CLIENT = None  # Initialize lazily when needed
GEMINI_EXCEPTIONS_TO_RETRY: Tuple[Type[Exception], ...] = (ResourceExhausted,)
logging.getLogger("google.genai").setLevel(logging.ERROR)
os.environ["GRPC_VERBOSITY"] = "ERROR"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
warnings.filterwarnings("ignore")

SAFETY_SETTINGS = {
    types.HarmCategory.HARM_CATEGORY_HATE_SPEECH: types.HarmBlockThreshold.BLOCK_NONE,
    types.HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: types.HarmBlockThreshold.BLOCK_NONE,
    types.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: types.HarmBlockThreshold.BLOCK_NONE,
    types.HarmCategory.HARM_CATEGORY_HARASSMENT: types.HarmBlockThreshold.BLOCK_NONE,
}


def _get_client():
    """Get or create the Gemini client lazily."""
    global _CLIENT
    if _CLIENT is None:
        _CLIENT = genai.Client()
    return _CLIENT


class GeminiAPI(VendorBase):
    used_for_structured_outputs: bool = True
    exceptions_to_retry: Tuple[Type[Exception], ...] = GEMINI_EXCEPTIONS_TO_RETRY

    def __init__(
        self,
        exceptions_to_retry: Tuple[Type[Exception], ...] = GEMINI_EXCEPTIONS_TO_RETRY,
        used_for_structured_outputs: bool = False,
    ):
        self.used_for_structured_outputs = used_for_structured_outputs
        self.exceptions_to_retry = exceptions_to_retry


    def get_aliased_model_name(self, model_name: str) -> str:
        if model_name in ALIASES:
            return ALIASES[model_name]
        return model_name

    @staticmethod
    def _msg_to_contents(messages: List[Dict[str, Any]]) -> List[types.Content]:
        # contents, sys_instr = [], None
        contents = []
        for m in messages:
            # if m["role"] == "system":
            #     sys_instr = f"<instructions>\n{m['content']}\n</instructions>"
            #     continue
            # text = (sys_instr + "\n" + m["content"]) if sys_instr else m["content"]
            if m["role"].lower() not in ["user", "assistant"]:
                continue
            role = "user" if m["role"] == "user" else "assistant"
            contents.append(types.Content(role=role, parts=[types.Part.from_text(text=m["content"])]))
        return contents

    @staticmethod
    def _tools_to_genai(tools: List[BaseTool]) -> List[types.Tool]:
        """Convert internal BaseTool → genai Tool."""
        out: List[types.Tool] = []
        for t in tools:
            # Assume t.to_gemini_tool() now correctly returns a FunctionDeclaration
            #func_decl = t.to_gemini_tool()
            if isinstance(t, dict):
                func_decl = t
            else:
                func_decl = t.to_gemini_tool()
            if not isinstance(func_decl, types.FunctionDeclaration):
                 # Or fetch schema parts if to_gemini_tool still returns dict
                 # This depends on BaseTool.to_gemini_tool implementation
                tool_dict = func_decl # Assuming it's a dict for now
                func_decl = types.FunctionDeclaration(
                    name=tool_dict['name'],
                    description=tool_dict['description'],
                    parameters=tool_dict['parameters'], # Expects OpenAPI-style dict
                )
            out.append(types.Tool(function_declarations=[func_decl]))
        return out

    async def _gen_content_async(
        self,
        messages: List[Dict],
        temperature: float,
        model_name: str,
        reasoning_effort: str,
        tools: Optional[List[BaseTool]],
        lm_config: Optional[Dict[str, Any]],
    ) -> Tuple[str, Optional[List[Dict]]]:
        model_name = self.get_aliased_model_name(model_name)
        cfg_kwargs: Dict[str, Any] = {"temperature": temperature}
        if model_name in GEMINI_REASONING_MODELS and reasoning_effort in GEMINI_THINKING_BUDGETS:
            cfg_kwargs["thinking_config"] = types.ThinkingConfig(
                thinking_budget=GEMINI_THINKING_BUDGETS[reasoning_effort]
            )
        
        if any(m["role"] == "system" for m in messages):
            cfg_kwargs["system_instruction"] = next(m["content"] for m in messages if m["role"] == "system")
        
        generation_config = types.GenerateContentConfig(
            **cfg_kwargs,
            tool_config=lm_config.get("tool_config") if lm_config else None,
            tools=self._tools_to_genai(tools) if tools else None
        )
        client = _get_client()
        resp = await client.aio.models.generate_content(
            model=model_name,
            contents=self._msg_to_contents(messages),
            config=generation_config,
            #safety_settings=SAFETY_SETTINGS,
        )
        return self._extract(resp)

    def _gen_content_sync(
        self,
        messages: List[Dict],
        temperature: float,
        model_name: str,
        reasoning_effort: str,
        tools: Optional[List[BaseTool]],
        lm_config: Optional[Dict[str, Any]],
    ) -> Tuple[str, Optional[List[Dict]]]:
        model_name = self.get_aliased_model_name(model_name)
        cfg_kwargs: Dict[str, Any] = {"temperature": temperature}
        if model_name in GEMINI_REASONING_MODELS and reasoning_effort in GEMINI_THINKING_BUDGETS:
            cfg_kwargs["thinking_config"] = types.ThinkingConfig(
                thinking_budget=GEMINI_THINKING_BUDGETS[reasoning_effort]
            )
        if any(m["role"] == "system" for m in messages):
            cfg_kwargs["system_instruction"] = next(m["content"] for m in messages if m["role"] == "system")
        generation_config = types.GenerateContentConfig(
            **cfg_kwargs,
            tool_config=lm_config.get("tool_config") if lm_config else None,
            tools=self._tools_to_genai(tools) if tools else None
        )

        client = _get_client()
        resp = client.models.generate_content(
            model=model_name,
            contents=self._msg_to_contents(messages),
            safety_settings=SAFETY_SETTINGS,
            config=generation_config,
        )
        return self._extract(resp)

    @staticmethod
    def _extract(response) -> Tuple[str, Optional[List[Dict]]]:
        # Extract text, handling cases where it might be missing
        try:
            text = response.text
        except ValueError: # Handle cases where only non-text parts exist
            text = "" 

        calls = []
        # Access parts through candidates[0].content
        if response.candidates and response.candidates[0].content:
            for part in response.candidates[0].content.parts:
                if part.function_call:
                    calls.append(
                        {
                            "id": f"call_{len(calls) + 1}",
                            "type": "function",
                            "function": {
                                "name": part.function_call.name,
                                "arguments": json.dumps(dict(part.function_call.args)),
                            },
                        }
                    )
        return text, calls or None

    @backoff.on_exception(
        backoff.expo,
        exceptions_to_retry,
        max_tries=BACKOFF_TOLERANCE,
        on_giveup=lambda e: print(e),
    )
    async def _hit_api_async(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        used_cache_handler = get_cache_handler(use_ephemeral_cache_only)
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        raw_response, tool_calls = await self._gen_content_async(
            messages,
            temperature=lm_config.get("temperature", SPECIAL_BASE_TEMPS.get(model, 0)),
            reasoning_effort=reasoning_effort,
            tools=tools,
            lm_config=lm_config,
            model_name=model,
        )
        if not raw_response:
            raw_response = ""
        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls,
        )

        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response

    @backoff.on_exception(
        backoff.expo,
        exceptions_to_retry,
        max_tries=BACKOFF_TOLERANCE,
        on_giveup=lambda e: print(e),
    )
    def _hit_api_sync(
        self,
        model: str,
        messages: List[Dict[str, Any]],
        lm_config: Dict[str, Any],
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
        tools: Optional[List[BaseTool]] = None,
    ) -> BaseLMResponse:
        assert (
            lm_config.get("response_model", None) is None
        ), "response_model is not supported for standard calls"
        used_cache_handler = get_cache_handler(
            use_ephemeral_cache_only=use_ephemeral_cache_only
        )
        lm_config["reasoning_effort"] = reasoning_effort
        cache_result = used_cache_handler.hit_managed_cache(
            model, messages, lm_config=lm_config, tools=tools
        )
        if cache_result:
            return cache_result

        raw_response, tool_calls = self._gen_content_sync(
            messages,
            temperature=lm_config.get("temperature", SPECIAL_BASE_TEMPS.get(model, 0)),
            reasoning_effort=reasoning_effort,
            tools=tools,
            lm_config=lm_config,
            model_name=model,
        )
        if not raw_response:
            raw_response = ""
        lm_response = BaseLMResponse(
            raw_response=raw_response,
            structured_output=None,
            tool_calls=tool_calls,
        )

        lm_config["reasoning_effort"] = reasoning_effort
        used_cache_handler.add_to_managed_cache(
            model, messages, lm_config=lm_config, output=lm_response, tools=tools
        )
        return lm_response



import asyncio
import unittest
from typing import List

from pydantic import BaseModel, Field

from synth_ai.zyk.lms.core.main import LM


# Define example structured output models
class SimpleResponse(BaseModel):
    message: str
    confidence_between_zero_one: float = Field(
        ..., description="Confidence level between 0 and 1"
    )


class ComplexResponse(BaseModel):
    title: str
    tags: List[str]
    content: str


class NestedResponse(BaseModel):
    main_category: str
    subcategories: List[str]
    details: SimpleResponse


class TestLMStructuredOutputs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Initialize LMs for both forced_json and stringified_json modes
        cls.lm_forced_json = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.7,
            max_retries="Few",
            structured_output_mode="forced_json",
        )
        cls.lm_stringified_json = LM(
            model_name="gpt-4o-mini",
            formatting_model_name="gpt-4o-mini",
            temperature=0.7,
            max_retries="Few",
            structured_output_mode="stringified_json",
        )

    def test_sync_simple_response(self):
        for lm in [self.lm_forced_json, self.lm_stringified_json]:
            with self.subTest(
                mode=lm.structured_output_handler.handler.structured_output_mode
            ):
                result = lm.respond_sync(
                    system_message="You are a helpful assistant.",
                    user_message="Give me a short greeting and your confidence level.",
                    response_model=SimpleResponse,
                )
                self.assertIsInstance(result, SimpleResponse)
                self.assertIsInstance(result.message, str)
                self.assertIsInstance(result.confidence_between_zero_one, float)
                self.assertGreaterEqual(result.confidence_between_zero_one, 0)
                self.assertLessEqual(result.confidence_between_zero_one, 1)

    def test_sync_complex_response(self):
        for lm in [self.lm_forced_json, self.lm_stringified_json]:
            with self.subTest(
                mode=lm.structured_output_handler.handler.structured_output_mode
            ):
                result = lm.respond_sync(
                    system_message="You are a content creator.",
                    user_message="Create a short blog post about AI.",
                    response_model=ComplexResponse,
                )
                self.assertIsInstance(result, ComplexResponse)
                self.assertIsInstance(result.title, str)
                self.assertIsInstance(result.tags, list)
                self.assertIsInstance(result.content, str)

    async def async_nested_response(self, lm):
        result = await lm.respond_async(
            system_message="You are a categorization expert.",
            user_message="Categorize 'Python' and provide a brief description.",
            response_model=NestedResponse,
        )
        self.assertIsInstance(result, NestedResponse)
        self.assertIsInstance(result.main_category, str)
        self.assertIsInstance(result.subcategories, list)
        self.assertIsInstance(result.details, SimpleResponse)

    def test_async_nested_response(self):
        for lm in [self.lm_forced_json, self.lm_stringified_json]:  #
            with self.subTest(
                mode=lm.structured_output_handler.handler.structured_output_mode
            ):
                asyncio.run(self.async_nested_response(lm))


if __name__ == "__main__":
    unittest.main()



OPENAI_REASONING_MODELS = ["o4", "o4-mini", "o3","o3-mini", "o1-mini", "o1"]
CLAUDE_REASONING_MODELS = ["claude-3-7-sonnet-latest"]
GEMINI_REASONING_MODELS = ["gemini-2.5-flash", "gemini-2.5-pro"]

# Gemini models that support thinking
GEMINI_REASONING_MODELS = ["gemini-2.5-flash", "gemini-2.5-pro"]
GEMINI_THINKING_BUDGETS = {
    "high": 10000,
    "medium": 5000,
    "low": 2500,
}

# Anthropic Sonnet 3.7 budgets
SONNET_37_BUDGETS = {
    "high": 8192,
    "medium": 4096,
    "low": 2048,
}

REASONING_MODELS = OPENAI_REASONING_MODELS + CLAUDE_REASONING_MODELS + GEMINI_REASONING_MODELS

SPECIAL_BASE_TEMPS = {model: 1 for model in REASONING_MODELS}


import os
from dataclasses import dataclass
from typing import Optional, Union

from diskcache import Cache
from pydantic import BaseModel

from synth_ai.zyk.lms.caching.constants import DISKCACHE_SIZE_LIMIT
from synth_ai.zyk.lms.vendors.base import BaseLMResponse


@dataclass
class EphemeralCache:
    def __init__(self, fast_cache_dir: str = ".cache/ephemeral_cache"):
        os.makedirs(fast_cache_dir, exist_ok=True)
        self.fast_cache = Cache(fast_cache_dir, size_limit=DISKCACHE_SIZE_LIMIT)

    def hit_cache(
        self, key: str, response_model: Optional[BaseModel] = None
    ) -> Optional[BaseLMResponse]:
        if key not in self.fast_cache:
            return None

        try:
            cache_data = self.fast_cache[key]
        except AttributeError:
            return None

        if not isinstance(cache_data, dict):
            return BaseLMResponse(
                raw_response=cache_data, structured_output=None, tool_calls=None
            )

        raw_response = cache_data.get("raw_response")
        tool_calls = cache_data.get("tool_calls")
        structured_output = cache_data.get("structured_output")

        if response_model and structured_output:
            structured_output = response_model(**structured_output)

        return BaseLMResponse(
            raw_response=raw_response,
            structured_output=structured_output,
            tool_calls=tool_calls,
        )

    def add_to_cache(self, key: str, response: Union[BaseLMResponse, str]) -> None:
        if isinstance(response, str):
            self.fast_cache[key] = response
            return

        if isinstance(response, BaseLMResponse):
            cache_data = {
                "raw_response": response.raw_response
                if response.raw_response is not None
                else None,
                "tool_calls": response.tool_calls
                if response.tool_calls is not None
                else None,
                "structured_output": (
                    response.structured_output.model_dump()
                    if response.structured_output is not None
                    else None
                ),
            }
            self.fast_cache[key] = cache_data
            return

        raise ValueError(f"Invalid response type: {type(response)}")

    def close(self):
        self.fast_cache.close()



from setuptools import find_packages, setup

setup(
    name="synth-ai",
    version="0.1.5",
    packages=find_packages(),
    install_requires=[
        "openai",
        "pydantic>=2.9.2",
        "diskcache",
        "backoff>=2.2.1",
        "anthropic>=0.34.2",
        "google>=3.0.0",
        "google-generativeai>=0.8.1",
        "together>=1.2.12",
    ],
    author="Josh Purtell",
    author_email="josh@usesynth.com",
    description="",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/synth-laboratories/synth-ai",
    license="MIT",
)


from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Literal, Optional, Union


class StructuredOutputCoercionFailureException(Exception):
    pass


# ... rest of imports ...



import logging
import time
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Literal, Optional, Union

from pydantic import BaseModel

from synth_ai.zyk.lms.core.exceptions import StructuredOutputCoercionFailureException
from synth_ai.zyk.lms.structured_outputs.inject import (
    inject_structured_output_instructions,
)
from synth_ai.zyk.lms.structured_outputs.rehabilitate import (
    fix_errant_forced_async,
    fix_errant_forced_sync,
    pull_out_structured_output,
)
from synth_ai.zyk.lms.vendors.base import BaseLMResponse, VendorBase
from synth_ai.zyk.lms.constants import SPECIAL_BASE_TEMPS

logger = logging.getLogger(__name__)


class StructuredHandlerBase(ABC):
    core_client: VendorBase
    retry_client: VendorBase
    handler_params: Dict[str, Any]
    structured_output_mode: Literal["stringified_json", "forced_json"]

    def __init__(
        self,
        core_client: VendorBase,
        retry_client: VendorBase,
        handler_params: Optional[Dict[str, Any]] = None,
        structured_output_mode: Literal[
            "stringified_json", "forced_json"
        ] = "stringified_json",
    ):
        self.core_client = core_client
        self.retry_client = retry_client
        self.handler_params = (
            handler_params if handler_params is not None else {"retries": 3}
        )
        self.structured_output_mode = structured_output_mode

    async def call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        temperature: float = 0.0,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        if temperature == 0.0:
            temperature = SPECIAL_BASE_TEMPS.get(model, 0.0)
        # print("Calling from base")
        return await self._process_call_async(
            messages=messages,
            model=model,
            response_model=response_model,
            api_call_method=self.core_client._hit_api_async_structured_output
            if (not not response_model and self.structured_output_mode == "forced_json")
            else self.core_client._hit_api_async,
            temperature=temperature,
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )

    def call_sync(
        self,
        messages: List[Dict[str, Any]],
        response_model: BaseModel,
        model: str,
        temperature: float = 0.0,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        if temperature == 0.0:
            temperature = SPECIAL_BASE_TEMPS.get(model, 0.0)
        return self._process_call_sync(
            messages=messages,
            model=model,
            response_model=response_model,
            api_call_method=self.core_client._hit_api_sync_structured_output
            if (not not response_model and self.structured_output_mode == "forced_json")
            else self.core_client._hit_api_sync,
            temperature=temperature,
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )

    @abstractmethod
    async def _process_call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        api_call_method,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        pass

    @abstractmethod
    def _process_call_sync(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        api_call_method,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        pass


class StringifiedJSONHandler(StructuredHandlerBase):
    core_client: VendorBase
    retry_client: VendorBase
    handler_params: Dict[str, Any]

    def __init__(
        self,
        core_client: VendorBase,
        retry_client: VendorBase,
        handler_params: Dict[str, Any] = {"retries": 3},
    ):
        super().__init__(
            core_client,
            retry_client,
            handler_params,
            structured_output_mode="stringified_json",
        )

    async def _process_call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        temperature: float,
        api_call_method: Callable,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        #ogger.info(f"Processing structured output call for model: {model}")
        assert callable(api_call_method), "api_call_method must be a callable"
        assert (
            response_model is not None
        ), "Don't use this handler for unstructured outputs"
        remaining_retries = self.handler_params.get("retries", 2)
        previously_failed_error_messages = []
        structured_output = None

        while remaining_retries > 0:
            messages_with_json_formatting_instructions = (
                inject_structured_output_instructions(
                    messages=messages,
                    response_model=response_model,
                    previously_failed_error_messages=previously_failed_error_messages,
                )
            )
            t0 = time.time()
            raw_text_response_or_cached_hit = await api_call_method(
                messages=messages_with_json_formatting_instructions,
                model=model,
                lm_config={"response_model": None, "temperature": temperature},
                use_ephemeral_cache_only=use_ephemeral_cache_only,
                reasoning_effort=reasoning_effort,
            )
            #logger.debug(f"Time to get response: {time.time() - t0:.2f}s")

            # Check if we got a cached BaseLMResponse
            assert (
                type(raw_text_response_or_cached_hit) in [str, BaseLMResponse]
            ), f"Expected str or BaseLMResponse, got {type(raw_text_response_or_cached_hit)}"
            if type(raw_text_response_or_cached_hit) == BaseLMResponse:
                #print("Got cached hit, returning directly")
                raw_text_response = raw_text_response_or_cached_hit.raw_response
            else:
                raw_text_response = raw_text_response_or_cached_hit
            #logger.debug(f"Raw response from model:\n{raw_text_response}")

            #print("Trying to parse structured output")
            try:
                structured_output = pull_out_structured_output(
                    raw_text_response, response_model
                )

                #print("Successfully parsed structured output on first attempt")
                break
            except Exception as e:
                logger.warning(f"Failed to parse structured output: {str(e)}")
                try:
                    print("Attempting to fix with forced JSON parser")
                    structured_output = await fix_errant_forced_async(
                        messages_with_json_formatting_instructions,
                        raw_text_response,
                        response_model,
                        "gpt-4o-mini",
                    )
                    assert isinstance(structured_output, BaseModel), "Structured output must be a Pydantic model"
                    assert not isinstance(structured_output, BaseLMResponse), "Got BaseLMResponse instead of Pydantic model"
                    print("Successfully fixed and parsed structured output")
                    break
                except Exception as e:
                    logger.error(f"Failed to fix structured output: {str(e)}")
                    previously_failed_error_messages.append(
                        f"Generated attempt and got error. Attempt:\n\n{raw_text_response}\n\nError:\n\n{e}"
                    )
                    remaining_retries -= 1
                    logger.warning(f"Retries remaining: {remaining_retries}")

        if structured_output is None:
            logger.error("Failed to get structured output after all retries")
            raise StructuredOutputCoercionFailureException(
                "Failed to get structured output"
            )
        #print("Successfully parsed structured output")
        #print(structured_output)
        assert isinstance(structured_output, BaseModel), "Structured output must be a Pydantic model"
        assert not isinstance(structured_output, BaseLMResponse),"Got BaseLMResponse instead of Pydantic model" 
        return BaseLMResponse(
            raw_response=raw_text_response,
            structured_output=structured_output,
            tool_calls=None,
        )

    def _process_call_sync(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        temperature: float,
        api_call_method: Callable,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        #logger.info(f"Processing structured output call for model: {model}")
        assert callable(api_call_method), "api_call_method must be a callable"
        assert (
            response_model is not None
        ), "Don't use this handler for unstructured outputs"
        remaining_retries = self.handler_params.get("retries", 2)
        previously_failed_error_messages = []
        structured_output = None

        while remaining_retries > 0:
            messages_with_json_formatting_instructions = (
                inject_structured_output_instructions(
                    messages=messages,
                    response_model=response_model,
                    previously_failed_error_messages=previously_failed_error_messages,
                )
            )
            t0 = time.time()
            raw_text_response_or_cached_hit = api_call_method(
                messages=messages_with_json_formatting_instructions,
                model=model,
                lm_config={"response_model": None, "temperature": temperature},
                use_ephemeral_cache_only=use_ephemeral_cache_only,
                reasoning_effort=reasoning_effort,
            )
            logger.debug(f"Time to get response: {time.time() - t0:.2f}s")

            # Check if we got a cached BaseLMResponse
            assert (
                type(raw_text_response_or_cached_hit) in [str, BaseLMResponse]
            ), f"Expected str or BaseLMResponse, got {type(raw_text_response_or_cached_hit)}"
            if type(raw_text_response_or_cached_hit) == BaseLMResponse:
                logger.info("Got cached hit, returning directly")
                raw_text_response = raw_text_response_or_cached_hit.raw_response
            else:   
                raw_text_response = raw_text_response_or_cached_hit
            #logger.debug(f"Raw response from model:\n{raw_text_response}")

            try:
                structured_output = pull_out_structured_output(
                    raw_text_response, response_model
                )
                #print("Successfully parsed structured output on first attempt")
                break
            except Exception as e:
                logger.warning(f"Failed to parse structured output: {str(e)}")
                try:
                    print("Attempting to fix with forced JSON parser")
                    structured_output = fix_errant_forced_sync(
                        raw_text_response, response_model, "gpt-4o-mini"
                    )
                    print("Successfully fixed and parsed structured output")
                    break
                except Exception as e:
                    logger.error(f"Failed to fix structured output: {str(e)}")
                    previously_failed_error_messages.append(
                        f"Generated attempt and got error. Attempt:\n\n{raw_text_response}\n\nError:\n\n{e}"
                    )
                    remaining_retries -= 1
                    logger.warning(f"Retries remaining: {remaining_retries}")

        #print("Successfully parsed structured output")
        #print(structured_output)
        if structured_output is None:
            logger.error("Failed to get structured output after all retries")
            raise StructuredOutputCoercionFailureException(
                "Failed to get structured output"
            )
        return BaseLMResponse(
            raw_response=raw_text_response,
            structured_output=structured_output,
            tool_calls=None,
        )


class ForcedJSONHandler(StructuredHandlerBase):
    core_client: VendorBase
    retry_client: VendorBase
    handler_params: Dict[str, Any]

    def __init__(
        self,
        core_client: VendorBase,
        retry_client: VendorBase,
        handler_params: Dict[str, Any] = {},
        reasoning_effort: str = "high",
    ):
        super().__init__(
            core_client,
            retry_client,
            handler_params,
            structured_output_mode="forced_json",
        )
        self.reasoning_effort = reasoning_effort

    async def _process_call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        api_call_method: Callable,
        temperature: float = 0.0,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        # print("Forced JSON")
        assert (
            response_model is not None
        ), "Don't use this handler for unstructured outputs"
        return await api_call_method(
            messages=messages,
            model=model,
            response_model=response_model,
            temperature=temperature,
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )

    def _process_call_sync(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        api_call_method: Callable,
        temperature: float = 0.0,
        use_ephemeral_cache_only: bool = False,
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        assert (
            response_model is not None
        ), "Don't use this handler for unstructured outputs"
        return api_call_method(
            messages=messages,
            model=model,
            response_model=response_model,
            temperature=temperature,
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )


class StructuredOutputHandler:
    handler: Union[StringifiedJSONHandler, ForcedJSONHandler]
    mode: Literal["stringified_json", "forced_json"]
    handler_params: Dict[str, Any]

    def __init__(
        self,
        core_client: VendorBase,
        retry_client: VendorBase,
        mode: Literal["stringified_json", "forced_json"],
        handler_params: Dict[str, Any] = {},
    ):
        self.mode = mode
        if self.mode == "stringified_json":
            self.handler = StringifiedJSONHandler(
                core_client, retry_client, handler_params
            )
        elif self.mode == "forced_json":
            # print("Forced JSON")
            self.handler = ForcedJSONHandler(core_client, retry_client, handler_params)
        else:
            raise ValueError(f"Invalid mode: {mode}")

    async def call_async(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        use_ephemeral_cache_only: bool = False,
        lm_config: Dict[str, Any] = {},
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        # print("Output handler call async")
        return await self.handler.call_async(
            messages=messages,
            model=model,
            response_model=response_model,
            temperature=lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0.0)
            ),
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )

    def call_sync(
        self,
        messages: List[Dict[str, Any]],
        model: str,
        response_model: BaseModel,
        use_ephemeral_cache_only: bool = False,
        lm_config: Dict[str, Any] = {},
        reasoning_effort: str = "high",
    ) -> BaseLMResponse:
        return self.handler.call_sync(
            messages=messages,
            model=model,
            response_model=response_model,
            temperature=lm_config.get(
                "temperature", SPECIAL_BASE_TEMPS.get(model, 0.0)
            ),
            use_ephemeral_cache_only=use_ephemeral_cache_only,
            reasoning_effort=reasoning_effort,
        )



import ast
import json
import logging
import re
from typing import Dict, List, Type, Union

from pydantic import BaseModel

from synth_ai.zyk.lms.vendors.base import VendorBase
from synth_ai.zyk.lms.vendors.core.openai_api import OpenAIStructuredOutputClient


def pull_out_structured_output(
    response_raw: str, response_model: Type[BaseModel]
) -> BaseModel:
    logger = logging.getLogger(__name__)
    #logger.debug(f"Raw response received: {response_raw}")

    assert isinstance(
        response_raw, str
    ), f"Response raw is not a string: {type(response_raw)}"

    # Use regex to extract JSON content within ```json ... ```
    json_pattern = re.compile(r"```json\s*(\{.*\})\s*```", re.DOTALL)
    match = json_pattern.search(response_raw)
    if match:
        response_prepared = match.group(1).strip()
    else:
        # Fallback to existing parsing if no code fencing is found
        if "```" in response_raw:
            response_prepared = response_raw.split("```")[1].strip()
        else:
            response_prepared = response_raw.strip()

    # Replace "null" with '"None"' if needed (ensure this aligns with your data)
    response_prepared = response_prepared.replace("null", '"None"')

    logger.debug(f"Prepared response for JSON parsing: {response_prepared}")

    try:
        response = json.loads(response_prepared)
        final = response_model(**response)
    except json.JSONDecodeError as e:
        # Attempt to parse using ast.literal_eval as a fallback
        response_prepared = response_prepared.replace("\n", "").replace("\\n", "")
        response_prepared = response_prepared.replace('\\"', '"')
        try:
            response = ast.literal_eval(response_prepared)
            final = response_model(**response)
        except Exception as inner_e:
            raise ValueError(
                f"Failed to parse response as {response_model}: {inner_e} - {response_prepared}"
            )
    except Exception as e:
        raise ValueError(
            f"Failed to parse response as {response_model}: {e} - {response_prepared}"
        )
    assert isinstance(final, BaseModel), "Structured output must be a Pydantic model"
    return final


def fix_errant_stringified_json_sync(
    response_raw: str,
    response_model: Type[BaseModel],
    models: List[str] = ["gpt-4o-mini", "gpt-4o"],
) -> BaseModel:
    try:
        return pull_out_structured_output(response_raw, response_model)
    except ValueError as e:
        # print("")
        mini_client = OpenAIStructuredOutputClient()
        messages = [
            {
                "role": "system",
                "content": "An AI attempted to generate stringified json that could be extracted into the provided Pydantic model. This json cannot be extracted, and an error message is provided to elucidate why. Please review the information and return a corrected response. Do not materially change the content of the response, just formatting where necessary.",
            },
            {
                "role": "user",
                "content": f"# Errant Attempt\n{response_raw}\n# Response Model\n {response_model.model_json_schema()}\n# Error Message\n {str(e)}.",
            },
        ]
        for model in models:
            if model == "gpt-4o":
                print(
                    "Warning - using gpt-4o for structured output correction - this could add up over time (latency, cost)"
                )
            try:
                fixed_response = mini_client._hit_api_sync(
                    model, messages=messages, lm_config={"temperature": 0.0}
                )
                return pull_out_structured_output(fixed_response, response_model)
            except Exception as e:
                pass
        raise ValueError("Failed to fix response using any model")


async def fix_errant_stringified_json_async(
    response_raw: str,
    response_model: Type[BaseModel],
    models: List[str] = ["gpt-4o-mini", "gpt-4o"],
) -> BaseModel:
    try:
        return pull_out_structured_output(response_raw, response_model)
    except ValueError as e:
        mini_client = OpenAIStructuredOutputClient()
        messages = [
            {
                "role": "system",
                "content": "An AI attempted to generate stringified json that could be extracted into the provided Pydantic model. This json cannot be extracted, and an error message is provided to elucidate why. Please review the information and return a corrected response. Do not materially change the content of the response, just formatting where necessary.",
            },
            {
                "role": "user",
                "content": f"# Errant Attempt\n{response_raw}\n# Response Model\n {response_model.model_json_schema()}\n# Error Message\n {str(e)}.",
            },
        ]
        for model in models:
            if model == "gpt-4o":
                print(
                    "Warning - using gpt-4o for structured output correction - this could add up over time (latency, cost)"
                )
            try:
                fixed_response = await mini_client._hit_api_async(
                    model, messages=messages, lm_config={"temperature": 0.0}
                )
                return pull_out_structured_output(fixed_response, response_model)
            except Exception as e:
                pass
        raise ValueError("Failed to fix response using any model")


async def fix_errant_forced_async(
    messages: List[Dict],
    response_raw: str,
    response_model: Type[BaseModel],
    model: str,
) -> BaseModel:
    try:
        return pull_out_structured_output(response_raw, response_model)
    except ValueError as e:
        client = OpenAIStructuredOutputClient()
        messages = [
            {
                "role": "system",
                "content": "An AI attempted to generate stringified json that could be extracted into the provided Pydantic model. This json cannot be extracted, and an error message is provided to elucidate why. Please review the information and return a corrected response. Do not materially change the content of the response, just formatting where necessary.",
            },
            {
                "role": "user",
                "content": f"<previous messages>\n<system_message>{messages[0]['content']}</system_message>\n<user_message>{messages[1]['content']}</user_message></previous messages> # Errant Attempt\n{response_raw}\n# Response Model\n {response_model.model_json_schema()}\n# Error Message\n {str(e)}.",
            },
        ]
        # print("Broken response:")
        # print(response_raw)
        fixed_response = await client._hit_api_async_structured_output(
            model=model,
            messages=messages,
            response_model=response_model,
            temperature=0.0,
        )
        # print("Fixed response:")
        # print(fixed_response)
        return fixed_response.structured_output


def fix_errant_forced_sync(
    response_raw: str,
    response_model: Type[BaseModel],
    model: str,
) -> BaseModel:
    client = OpenAIStructuredOutputClient()
    messages = [
        {
            "role": "system",
            "content": "An AI attempted to generate a response that could be extracted into the provided Pydantic model. This response cannot be extracted. Please review the information and return a corrected response.",
        },
        {
            "role": "user",
            "content": f"# Errant Attempt\n{response_raw}\n# Response Model\n {response_model.model_json_schema()}",
        },
    ]
    # print("Broken response:")
    # print(response_raw)
    fixed_response = client._hit_api_sync_structured_output(
        model=model, messages=messages, response_model=response_model, temperature=0.0
    )
    # print("Fixed response:")
    # print(fixed_response)
    return fixed_response.structured_output
